from amuse.support import interface
from amuse.support.exceptions import AmuseException


from amuse.datamodel.binding import *
from amuse.datamodel.parameters import *
from amuse.support.core import OrderedDictionary
from amuse.support import exceptions

from amuse.test import amusetest
import numpy
import pickle
from amuse.units import units
from amuse.units import nbody_system
from amuse import datamodel
from amuse.rfi.core import legacy_function
from amuse.rfi.core import LegacyFunctionSpecification


class CodeInterfaceWithConvertedUnitsTests(amusetest.TestCase):
    class TestClass(object):

        def get_mass(self):
            return 10.0, 0

        def add_to_length(self, length):
            return length + 10.0

        def return_an_errorcode(self, error):
            return error

    def test1(self):
        convert_nbody = nbody_system.nbody_to_si(10.0 | units.kg, 5.0 | units.m)

        original = self.TestClass()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('PROPERTY')
        handler.add_property('get_mass', nbody_system.mass)

        handler = instance.get_handler('UNIT')
        handler.set_nbody_converter(convert_nbody)

        handler = instance.get_handler('METHOD')
        handler.add_method('add_to_length', (nbody_system.length,), nbody_system.length)

        self.assertAlmostEqual(instance.mass.value_in(units.kg), 100.0, 10)

    def test2(self):
        convert_nbody = nbody_system.nbody_to_si(10.0 | units.kg, 5.0 | units.m)

        original = self.TestClass()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('PROPERTY')
        handler.add_property('get_mass', nbody_system.mass)

        handler = instance.get_handler('UNIT')
        handler.set_nbody_converter(convert_nbody)

        handler = instance.get_handler('METHOD')
        handler.add_method('add_to_length', (nbody_system.length,), nbody_system.length)

        self.assertAlmostEqual(instance.add_to_length(5 | units.m).value_in(units.m), 55.0, 10)

    def test3(self):
        original = self.TestClass()
        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('METHOD')
        handler.add_method('add_to_length', (units.m,), units.m, public_name='add_10')

        self.assertFalse(instance.add_10.is_async_supported)

    def test4(self):
        original = self.TestClass()
        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('METHOD')
        handler.add_method('return_an_errorcode', (handler.NO_UNIT,), handler.ERROR_CODE)

        self.assertEqual(instance.return_an_errorcode(0), None)
        self.assertRaises(Exception, lambda: instance.return_an_errorcode(-1),
            expected_message="Error when calling 'return_an_errorcode' of a '<class 'amuse.support.interface.InCodeComponentImplementation'>', errorcode is -1"
        )


class CodeInterfaceWithUnitsOnLegacyFunctionTests(amusetest.TestCase):

    def test1(self):
        class TestImplementation(object):

            def get_mass(self):
                return 10.0, 0

        class TestInterface(object):

            @legacy_function
            def get_mass():
                function = LegacyFunctionSpecification()
                function.addParameter('input1', dtype='d', direction=function.OUT, unit=units.kg)
                function.result_type = 'i'
                return function

        original = TestInterface()
        instance = interface.InCodeComponentImplementation(original)

        instance.get_handler("LEGACY").legacy_interface = TestImplementation()
        self.assertAlmostRelativeEquals(instance.get_mass(), 10 | units.kg)

    def test2(self):
        class TestImplementation(object):

            def echo_one(self, input):
                return input, 0

        class TestInterface(object):

            @legacy_function
            def echo_one():
                function = LegacyFunctionSpecification()
                function.addParameter('input', dtype='d', direction=function.IN, unit=units.kg)
                function.addParameter('output', dtype='d', direction=function.OUT, unit=units.g)
                function.result_type = 'i'
                return function

        original = TestInterface()
        instance = interface.InCodeComponentImplementation(original)

        instance.get_handler("LEGACY").legacy_interface = TestImplementation()
        self.assertAlmostRelativeEquals(instance.echo_one(1 | units.kg), 1 | units.g)

    def test3(self):
        class TestImplementation(object):

            def return_error(self):
                return -1

        class TestInterface(object):

            @legacy_function
            def return_error():
                function = LegacyFunctionSpecification()
                function.result_type = 'i'
                return function

        original = TestInterface()
        instance = interface.InCodeComponentImplementation(original)

        instance.get_handler("LEGACY").legacy_interface = TestImplementation()
        self.assertRaises(Exception, instance.return_error, " Error when calling 'echo_one' of a '<class 'amuse.support.interface.InCodeComponentImplementation'>', errorcode is -1")

    def test4(self):
        class TestImplementation(object):

            def echo_one(self, input):
                return input, 0

        class TestInterface(object):

            @legacy_function
            def echo_one():
                function = LegacyFunctionSpecification()
                function.addParameter('input', dtype='d', direction=function.IN, unit=units.deg)
                function.addParameter('output', dtype='d', direction=function.OUT, unit=units.deg)
                function.result_type = 'i'
                return function

        original = TestInterface()
        instance = interface.InCodeComponentImplementation(original)

        instance.get_handler("LEGACY").legacy_interface = TestImplementation()
        self.assertAlmostRelativeEquals(instance.echo_one(1. | units.rad), 1. | units.rad)
        # this was: self.assertAlmostRelativeEquals(instance.echo_one(1.), 1. | units.deg)
        # but after 85bd5d99 or 945bc46, it needs to be:
        self.assertAlmostRelativeEquals(instance.echo_one(1.), 1. | units.rad)
        # this is indeed proper behaviour!!


class CodeInterfaceWithMethodsAndPropertiesTests(amusetest.TestCase):
    class TestClass(object):

        def add_10_to_length(self, length):
            return length + 10

        def get_one(self):
            return 1.0, 0.0

        def get_state(self, id):
            return (1.0, 2.0, 3.0, 0.0)

        def get_state_error(self, id):
            return (1.0, 2.0, 3.0, -1.0)

    def test1(self):
        original = self.TestClass()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('METHOD')
        handler.add_method('add_10_to_length', (units.m,), units.m)

        self.assertEqual(20.0, original.add_10_to_length(10.0))

        self.assertEqual(20.0 | units.m, instance.add_10_to_length(10.0 | units.m))
        self.assertEqual(1010.0 | units.m, instance.add_10_to_length(1.0 | units.km))

    def test2(self):
        original = self.TestClass()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('METHOD')
        handler.add_method('add_10_to_length', (units.m,), units.m, public_name='add_10')

        self.assertEqual(20.0 | units.m, instance.add_10(10.0 | units.m))
        self.assertEqual(1010.0 | units.m, instance.add_10(1.0 | units.km))

    def test3(self):
        original = self.TestClass()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('PROPERTY')
        handler.add_property('get_one', units.m)

        self.assertEqual(1.0 | units.m, instance.one)

    def test4(self):
        original = self.TestClass()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('PROPERTY')
        handler.add_property('get_one', units.m, public_name='get_one')

        self.assertEqual(1.0 | units.m, instance.get_one)

    def test5(self):
        original = self.TestClass()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('METHOD')
        handler.add_method('get_state', (handler.NO_UNIT,), (units.m, units.m, units.kg, handler.ERROR_CODE))

        result = instance.get_state(1)
        self.assertEqual(3, len(result))
        self.assertEqual(1.0 | units.m, result[0])
        self.assertEqual(3.0 | units.kg, result[2])

    def test6(self):
        original = self.TestClass()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('METHOD')
        handler.add_method('get_state_error', (handler.NO_UNIT,), (units.m, units.m, units.kg, handler.ERROR_CODE))

        self.assertRaises(AmuseException, instance.get_state_error, 1,
            expected_message="Error when calling 'get_state_error' of a '<class 'amuse.support.interface.InCodeComponentImplementation'>', errorcode is -1.0")


class ClassWithState(object):

    def __init__(self):
        self.state = 0
        self.number_of_times_move_to_state_1_called = 0
        self.number_of_times_move_to_state_2_called = 0
        self.number_of_times_move_to_state_3_called = 0
        self.number_of_times_move_to_state_4_called = 0

    def always_works(self):
        return True

    def move_to_state_1(self):
        self.number_of_times_move_to_state_1_called += 1
        self.state = 1

    def move_to_state_2(self):
        self.number_of_times_move_to_state_2_called += 1
        self.state = 2

    def move_to_state_3(self):
        self.number_of_times_move_to_state_3_called += 1
        self.state = 3

    def move_to_state_4(self):
        self.number_of_times_move_to_state_4_called += 1
        self.state = 4

    def returns_1(self):
        return self.state

    def returns_2(self):
        return self.state

    def returns_3(self):
        return self.state

    def returns_4(self):
        return self.state


class CodeInterfaceTests(amusetest.TestCase):

    def test1(self):
        original = ClassWithState()

        instance = interface.InCodeComponentImplementation(original)

        self.assertTrue(instance.always_works())
        instance.move_to_state_1()
        self.assertEqual(1, instance.returns_1())

    def test2(self):
        original = ClassWithState()

        instance = interface.InCodeComponentImplementation(original)
        handler = instance.get_handler('STATE')

        handler.add_transition('ZERO', 'ONE', 'move_to_state_1')
        handler.add_method('ONE', 'returns_1')
        handler.set_initial_state('ZERO')

        self.assertTrue(instance.always_works())
        self.assertEqual(instance.get_name_of_current_state(), 'ZERO')
        instance.move_to_state_1()

        self.assertEqual(instance.get_name_of_current_state(), 'ONE')
        self.assertEqual(instance.returns_1(), 1)

    def test3(self):
        class TestCodeInterface(interface.InCodeComponentImplementation):

            def move_to_state_1(self):
                self.overridden().move_to_state_1()
                self.traced = True

        original = ClassWithState()

        instance = TestCodeInterface(original)
        instance.traced = False

        handler = instance.get_handler('STATE')
        handler.add_transition('ZERO', 'ONE', 'move_to_state_1')
        handler.add_method('ONE', 'returns_1')
        handler.set_initial_state('ZERO')

        self.assertEqual(instance.get_name_of_current_state(), 'ZERO')
        self.assertEqual(instance.returns_1(), 1)
        self.assertEqual(instance.get_name_of_current_state(), 'ONE')
        self.assertTrue(instance.traced)

    def test4(self):
        original = ClassWithState()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('STATE')

        handler.add_transition('ZERO', 'ONE', 'move_to_state_1')
        handler.add_method('ONE', 'returns_1')
        handler.set_initial_state('ZERO')

        self.assertEqual(instance.get_name_of_current_state(), 'ZERO')
        self.assertEqual(instance.returns_1(), 1)
        self.assertEqual(instance.get_name_of_current_state(), 'ONE')

    def test5(self):
        original = ClassWithState()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('STATE')
        handler.add_transition('ZERO', 'ONE', 'move_to_state_1')
        handler.add_transition('ONE', 'TWO', 'move_to_state_2')
        handler.add_method('ONE', 'returns_1')
        handler.add_method('TWO', 'returns_2')
        handler.set_initial_state('ZERO')

        self.assertEqual(instance.get_name_of_current_state(), 'ZERO')
        self.assertEqual(instance.returns_2(), 2)
        self.assertEqual(instance.get_name_of_current_state(), 'TWO')

    def test6(self):
        original = ClassWithState()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('STATE')
        handler.add_transition('ZERO', 'ONE', 'move_to_state_1')
        handler.add_transition('ONE', 'TWO', 'move_to_state_2')
        handler.add_transition('TWO', 'THREE', 'move_to_state_3')
        handler.add_transition('TWO', 'FOUR', 'move_to_state_4')
        handler.add_transition('THREE', 'ONE', 'move_to_state_1')

        handler.add_method('ONE', 'returns_1')
        handler.add_method('TWO', 'returns_2')
        handler.add_method('THREE', 'returns_3')
        handler.add_method('FOUR', 'returns_4')
        handler.set_initial_state('ZERO')

        self.assertEqual(instance.get_name_of_current_state(), 'ZERO')
        self.assertEqual(instance.returns_4(), 4)
        self.assertEqual(instance.get_name_of_current_state(), 'FOUR')

        handler.set_initial_state('ZERO')
        self.assertEqual(instance.get_name_of_current_state(), 'ZERO')
        self.assertEqual(instance.returns_3(), 3)
        self.assertEqual(instance.get_name_of_current_state(), 'THREE')
        self.assertEqual(instance.returns_4(), 4)
        self.assertEqual(instance.get_name_of_current_state(), 'FOUR')

    def test7(self):
        original = ClassWithState()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('STATE')
        handler.add_transition('ZERO', 'ONE', 'move_to_state_1')
        handler.add_transition('ONE', 'TWO', 'move_to_state_2')
        handler.add_transition('TWO', 'THREE', 'move_to_state_3')
        handler.add_transition('TWO', 'FOUR', 'move_to_state_4')
        handler.add_transition('THREE', 'ONE', 'move_to_state_1')

        handler.add_method('ONE', 'returns_1')
        handler.add_method('TWO', 'returns_2')
        handler.add_method('THREE', 'returns_3')
        handler.add_method('FOUR', 'returns_4')
        handler.set_initial_state('ZERO')
        handler.do_automatic_state_transitions(False)

        self.assertEqual(instance.get_name_of_current_state(), 'ZERO')
        try:
            self.assertEqual(instance.returns_4(), 4)
            self.fail("Automatic state transitions is OFF, this method should fail")
        except Exception as ex:
            print(ex)

            self.assertEqual(len(ex.transitions), 3)

            for x in ex.transitions:
                x.do()

            self.assertEqual(instance.returns_4(), 4)

    def test8(self):
        original = ClassWithState()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('STATE')
        handler.add_transition('ZERO', 'ONE', 'move_to_state_1')
        handler.add_transition('ONE', 'TWO', 'move_to_state_2')

        handler.add_method('ONE', 'returns_1')
        handler.add_method('TWO', 'returns_2')
        handler.add_method('THREE', 'returns_3')
        handler.set_initial_state('ZERO')

        self.assertEqual(instance.get_name_of_current_state(), 'ZERO')
        self.assertEqual(instance.returns_2(), 2)
        self.assertRaises(Exception, instance.returns_3,
            expected_message="While calling returns_3 of <class 'amuse.support.interface.InCodeComponentImplementation'>: No transition from current state state 'TWO' to state 'THREE' possible")

    def test9(self):
        original = ClassWithState()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('STATE')
        handler.add_transition('ZERO', 'ONE', 'move_to_state_1')
        handler.add_transition('ONE', 'TWO', 'move_to_state_2')
        handler.add_transition('THREE', 'ONE', 'move_to_state_1')
        handler.add_transition('TWO', 'THREE', 'move_to_state_1')

        handler.add_method('ONE', 'returns_1')
        handler.add_method('TWO', 'returns_2')
        handler.add_method('THREE', 'returns_3')
        handler.set_initial_state('ZERO')

        self.assertEqual(instance.get_name_of_current_state(), 'ZERO')
        instance.move_to_state_1()
        self.assertEqual(instance.get_name_of_current_state(), 'ONE')
        instance.move_to_state_2()
        self.assertEqual(instance.get_name_of_current_state(), 'TWO')
        instance.move_to_state_1()
        self.assertEqual(instance.get_name_of_current_state(), 'THREE')
        instance.move_to_state_1()
        self.assertEqual(instance.get_name_of_current_state(), 'ONE')

        self.assertEqual(instance.returns_3(), 1)

    def test10(self):
        original = ClassWithState()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('STATE')
        handler.add_transition('ZERO', 'ONE', 'move_to_state_1')
        handler.add_transition('ONE', 'TWO', 'move_to_state_2')
        handler.add_transition('THREE', 'ONE', 'move_to_state_1')
        handler.add_transition('TWO', 'THREE', 'move_to_state_1')

        handler.add_method('ONE', 'returns_1')
        handler.add_method('TWO', 'returns_2')
        handler.add_method('THREE', 'returns_3')
        handler.set_initial_state('ZERO')

        self.assertTrue(instance.state_machine.is_enabled)
        self.assertEqual(instance.state_machine.get_name_of_current_state(), 'ZERO')
        instance.move_to_state_1()
        self.assertEqual(instance.get_name_of_current_state(), 'ONE')
        instance.state_machine.disable()
        self.assertFalse(instance.state_machine.is_enabled)
        instance.move_to_state_2()
        self.assertEqual(instance.get_name_of_current_state(), 'ONE')

    def test11(self):
        original = ClassWithState()

        instance = interface.InCodeComponentImplementation(original)
        handler = instance.get_handler('STATE')

        handler.add_transition('ZERO', 'ONE', 'move_to_state_1')
        handler.add_transition('!ZERO', 'TWO', 'move_to_state_2')
        handler.add_transition('TWO', 'ONE', 'move_to_state_1')
        handler.add_method('TWO', 'returns_1')
        handler.set_initial_state('ZERO')

        self.assertEqual(instance.get_name_of_current_state(), 'ZERO')
        instance.returns_1()

        self.assertEqual(instance.get_name_of_current_state(), 'TWO')
        self.assertEqual(instance.number_of_times_move_to_state_1_called, 1)
        self.assertEqual(instance.number_of_times_move_to_state_2_called, 1)

    def test12(self):
        original = ClassWithState()

        instance = interface.InCodeComponentImplementation(original, log_transitions=True)
        handler = instance.get_handler('STATE')
        handler.add_transition('ZERO', 'ONE', 'move_to_state_1')
        handler.add_transition('ONE', 'THREE', 'move_to_state_3')
        handler.add_transition('THREE', 'FOUR', 'move_to_state_4')
        handler.add_transition('!ZERO!ONE!THREE', 'TWO', 'move_to_state_2')
        handler.add_transition('TWO', 'ONE', 'move_to_state_1')
        handler.add_method('TWO', 'returns_1')
        handler.set_initial_state('ZERO')

        self.assertEqual(instance.get_name_of_current_state(), 'ZERO')
        instance.returns_1()

        self.assertEqual(instance.get_name_of_current_state(), 'TWO')
        self.assertEqual(instance.number_of_times_move_to_state_1_called, 1)
        self.assertEqual(instance.number_of_times_move_to_state_2_called, 1)
        self.assertEqual(instance.number_of_times_move_to_state_3_called, 1)
        self.assertEqual(instance.number_of_times_move_to_state_4_called, 1)

    def test13(self):
        original = ClassWithState()

        instance = interface.InCodeComponentImplementation(original, log_transitions=True)
        handler = instance.get_handler('STATE')
        handler.add_transition('!ZERO!ONE!TWO', 'FOUR', 'move_to_state_4')
        handler.add_transition('ONE', 'TWO', 'move_to_state_2')
        handler.add_transition('TWO', 'THREE', 'move_to_state_3')
        handler.add_transition('ZERO', 'ONE', 'move_to_state_1')
        handler.add_method('!ZERO', 'returns_1')
        handler.set_initial_state('ZERO')

        self.assertEqual(instance.get_name_of_current_state(), 'ZERO')
        instance.returns_1()

        self.assertEqual(instance.get_name_of_current_state(), 'ONE')
        self.assertEqual(instance.number_of_times_move_to_state_1_called, 1)
        self.assertEqual(instance.number_of_times_move_to_state_2_called, 0)
        self.assertEqual(instance.number_of_times_move_to_state_3_called, 0)
        self.assertEqual(instance.number_of_times_move_to_state_4_called, 0)
        instance.move_to_state_2()
        self.assertEqual(instance.get_name_of_current_state(), 'TWO')
        instance.returns_1()

        self.assertEqual(instance.get_name_of_current_state(), 'TWO')

    def test14(self):
        original = ClassWithState()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('STATE')
        handler.add_transition('ZERO', 'ONE', 'move_to_state_1')
        handler.add_transition('ONE', 'TWO', 'move_to_state_2')

        handler.add_method('ONE', 'returns_1')
        handler.add_method('TWO', 'returns_2')
        handler.set_initial_state('ZERO')
        handler.remove_transition('ZERO', 'ONE', 'move_to_state_1')
        handler.remove_transition('ONE', 'TWO', 'move_to_state_2')
        handler.add_transition('ZERO', 'TWO', 'move_to_state_1')

        self.assertEqual(instance.get_name_of_current_state(), 'ZERO')
        instance.returns_2()
        self.assertEqual(instance.get_name_of_current_state(), 'TWO')

    def test15(self):
        original = ClassWithState()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('STATE')
        handler.add_transition('ZERO', 'ONE', 'move_to_state_1')
        handler.add_transition('ONE', 'TWO', 'move_to_state_2')
        handler.add_transition('TWO', 'THREE', 'move_to_state_3')
        handler.add_transition('TWO', 'FOUR', 'move_to_state_4')
        handler.add_transition('THREE', 'ONE', 'move_to_state_1')

        handler.add_method('THREE', 'returns_1')
        handler.add_method('TWO', 'returns_1')
        handler.add_method('ONE', 'returns_1')
        handler.set_initial_state('ZERO')

        self.assertEqual(instance.get_name_of_current_state(), 'ZERO')
        self.assertEqual(instance.returns_1(), 1)
        self.assertEqual(instance.get_name_of_current_state(), 'ONE')

    def test9(self):
        original = ClassWithState()
        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('STATE')
        handler.add_transition('ZERO', 'ONE', 'move_to_state_1')
        handler.add_transition('ONE', 'TWO', 'move_to_state_2')
        handler.add_transition('TWO', 'THREE', 'move_to_state_3')
        handler.add_transition('TWO', 'FOUR', 'move_to_state_4')
        handler.add_transition('THREE', 'ONE', 'move_to_state_1')

        handler.add_method('THREE', 'returns_3')
        handler.add_method('TWO', 'returns_2')
        handler.add_method('ONE', 'returns_1')
        handler.set_initial_state('ZERO')

        self.assertEqual(instance.get_name_of_current_state(), 'ZERO')
        self.assertEqual(instance.returns_1(), 1)
        self.assertEqual(instance.get_name_of_current_state(), 'ONE')

        pickled_instance = pickle.dumps(instance)
        unpickled_instance = pickle.loads(pickled_instance)
        self.assertEqual(unpickled_instance.get_name_of_current_state(), 'ONE')
        self.assertEqual(unpickled_instance.returns_2(), 2)
        self.assertEqual(unpickled_instance.get_name_of_current_state(), 'TWO')
        self.assertEqual(instance.get_name_of_current_state(), 'ONE')

    def test10(self):
        original = ClassWithState()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('STATE')
        handler.add_transition('ZERO', 'ONE', 'move_to_state_1')
# remove to generate exception
#        handler.add_transition('ONE', 'TWO', 'move_to_state_2')
        handler.add_method('ONE', 'returns_1')
        handler.add_method('TWO', 'returns_2')
        handler.set_initial_state('ZERO')

        self.assertEqual(instance.get_name_of_current_state(), 'ZERO')
        self.assertRaises(Exception, instance.returns_2, expected_message="While calling returns_2 of <class 'amuse.support.interface.InCodeComponentImplementation'>: No transition from current state state 'ZERO' to state 'TWO' possible")


class CodeInterfaceWithUnitsAndStateTests(amusetest.TestCase):
    class TestClass(object):

        def __init__(self):
            self.value = 10.0

        def add_to_length(self, length):
            return length + self.value

        def move_to_20(self):
            self.value = 20

    def test1(self):
        original = self.TestClass()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('METHOD')
        handler.add_method('add_to_length', (units.m,), units.m)

        handler = instance.get_handler('STATE')
        handler.set_initial_state('ZERO')
        handler.add_transition('ZERO', 'ONE', 'move_to_20')
        handler.add_method('ONE', 'add_to_length')

        self.assertEqual(40.0 | units.m, instance.add_to_length(20.0 | units.m))

    def test2(self):
        original = self.TestClass()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('METHOD')
        handler.add_method('add_to_length', (units.m,), units.m, public_name='add')

        handler = instance.get_handler('STATE')
        handler.set_initial_state('ZERO')
        handler.add_transition('ZERO', 'ONE', 'move_to_20')
        handler.add_method('ONE', 'add_to_length')

        self.assertEqual(40.0 | units.m, instance.add(20.0 | units.m))

    def test3(self):
        original = self.TestClass()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('METHOD')
        handler.add_method('add_to_length', (units.m,), units.m, public_name='add')

        handler = instance.get_handler('STATE')
        handler.set_initial_state('ZERO')
        handler.add_transition('ZERO', 'ONE', 'move_to_20')
        handler.add_method('ONE', 'add')

        self.assertEqual(40.0 | units.m, instance.add(20.0 | units.m))


class CodeInterfaceWithErrorHandlingTests(amusetest.TestCase):
    class TestClass(object):
        errorcode = 0

        def get_mass(self):
            return 10.0, self.errorcode

    def test1(self):
        convert_nbody = nbody_system.nbody_to_si(10.0 | units.kg, 5.0 | units.m)

        original = self.TestClass()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('METHOD')
        handler.add_method('get_mass', (), (units.m, handler.ERROR_CODE,))
        handler = instance.get_handler('ERRORCODE')
        handler.add_errorcode(-2, "no such method")
        handler.add_errorcode(-3, "not available")

        self.assertEqual(instance.get_mass(), 10.0 | units.m)
        original.errorcode = -2
        self.assertRaises(AmuseException, instance.get_mass, expected_message="Error when calling 'get_mass' of a '<class 'amuse.support.interface.InCodeComponentImplementation'>', errorcode is -2, error is 'no such method'")

        original.errorcode = -1
        self.assertRaises(AmuseException, instance.get_mass, expected_message="Error when calling 'get_mass' of a '<class 'amuse.support.interface.InCodeComponentImplementation'>', errorcode is -1")


class CodeInterfaceWithParticlesTests(amusetest.TestCase):
    class TestClass(object):

        def get_mass(self):
            return 10.0, 0

        def add_to_length(self, length):
            return length + 10.0

    def test1(self):
        convert_nbody = nbody_system.nbody_to_si(10.0 | units.kg, 5.0 | units.m)

        original = self.TestClass()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('PROPERTY')
        handler.add_property('get_mass', nbody_system.mass)

        handler = instance.get_handler('UNIT')
        handler.set_nbody_converter(convert_nbody)


class ParticlesWithBindingInterface(object):

    def __init__(self):
        self.masses = {}

    def get_mass(self, id):
        masses = []
        errors = []
        for x in id:
            masses.append(self.masses[x])
            errors.append(0)
        result = OrderedDictionary()
        result["mass"] = masses
        result["__result"] = errors
        return result

    def set_mass(self, id, mass):
        for i, m in zip(id, mass):
            self.masses[i] = m

        return ([0] * len(id),)

    def new_particle(self, mass):
        ids = []
        errors = []

        for x in mass:
            id = len(self.masses)
            self.masses[len(self.masses)] = x
            ids.append(id)
            errors.append(0)

        return (ids, errors)

    def delete_particle(self, id):
        errors = []
        for x in id:
            self.masses[x].stop()
            errors.append(0)
        return errors

    def get_number_of_particles(self):
        return (len(self.masses), 0)

    def get_colliding_indices(self):
        return (1, 2, 0)

    def get_next(self, id):
        return ([x+1 for x in id], [0 for x in id])

    def add_1_to_mass(self, id):
        if isinstance(id, (int, numpy.int64, numpy.int32)):
            self.masses[id] += 1.0
            return [0]
        for i in id:
            self.masses[i] += 1.0
        return [0 for x in id]

    def get_heaviest_particle(self):
        max = -1
        id = -1
        for index, mass in self.masses.items():
            if mass > max:
                id = index
                max = mass
        return id

    def get_number_of_lightest_particles(self):
        return 2

    def get_lightest_particles(self, offset):
        sorted_masses = sorted(list(self.masses.items()), key=lambda x: x[1])
        return [sorted_masses[x][0] for x in offset]

    def get_number_of_particles_with_same_mass(self, id):
        result = []
        for index in id:
            mass = self.masses[index]
            count = 0

            for otherindex, m in self.masses.items():
                if otherindex == index:
                    continue
                if abs(m - mass) < 1.0:
                    count += 1
            result.append(count)
        return result

    def get_particle_with_same_mass(self, id, offset):
        result = []
        for index, index_in_list in zip(id, offset):
            mass = self.masses[index]
            count = 0
            for otherindex, m in self.masses.items():
                if otherindex == index:
                    continue
                if abs(m - mass) < 1.0:
                    count += 1
                    if count == (index_in_list + 1):
                        result.append(otherindex)
                        break
        return result


class TestParticlesWithBinding(amusetest.TestCase):

    def test1(self):
        original = ParticlesWithBindingInterface()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('METHOD')
        handler.add_method('get_mass', (handler.NO_UNIT,), (units.kg, handler.ERROR_CODE))
        handler.add_method('set_mass', (handler.NO_UNIT, units.kg,), (handler.ERROR_CODE,))
        handler.add_method('new_particle', (units.kg,), (handler.NO_UNIT, handler.ERROR_CODE))
        handler.add_method('delete_particle', (handler.NO_UNIT,), (handler.ERROR_CODE,))
        handler.add_method('get_number_of_particles', (), (handler.NO_UNIT, handler.ERROR_CODE,))

        handler = instance.get_handler('PARTICLES')
        handler.define_set('particles', 'id')
        handler.set_new('particles', 'new_particle')
        handler.set_delete('particles', 'delete_particle')
        handler.add_setter('particles', 'set_mass')
        handler.add_getter('particles', 'get_mass', names=('mass',))

        local_particles = datamodel.Particles(4)
        local_particles.mass = units.kg.new_quantity([3.0, 4.0, 5.0, 6.0])

        remote_particles = instance.particles
        remote_particles.add_particles(local_particles)

        self.assertEqual(len(original.masses), 4)
        self.assertEqual(original.masses[0], 3.0)
        self.assertEqual(original.masses[3], 6.0)

        self.assertEqual(len(instance.particles), 4)
        self.assertEqual(instance.particles[0].mass, 3.0 | units.kg)

    def test2(self):
        original = ParticlesWithBindingInterface()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('METHOD')
        handler.add_method('get_mass', (handler.NO_UNIT,), (units.kg, handler.ERROR_CODE))
        handler.add_method('set_mass', (handler.NO_UNIT, units.kg,), (handler.ERROR_CODE,))
        handler.add_method('new_particle', (units.kg,), (handler.NO_UNIT, handler.ERROR_CODE))
        handler.add_method('delete_particle', (handler.NO_UNIT,), (handler.ERROR_CODE,))
        handler.add_method('get_number_of_particles', (), (handler.NO_UNIT, handler.ERROR_CODE,))
        handler.add_method('get_colliding_indices', (), (handler.NO_UNIT, handler.NO_UNIT, handler.ERROR_CODE,))
        handler = instance.get_handler('PARTICLES')
        handler.define_set('particles', 'id')
        handler.set_new('particles', 'new_particle')
        handler.set_delete('particles', 'delete_particle')
        handler.add_setter('particles', 'set_mass')
        handler.add_getter('particles', 'get_mass', names=('mass',))

        handler.add_query('particles', 'get_colliding_indices', public_name='select_colliding')
        handler.add_select_from_particle('particles', 'get_next', public_name='next')

        local_particles = datamodel.Particles(4)
        local_particles.mass = units.kg.new_quantity([3.0, 4.0, 5.0, 6.0])

        remote_particles = instance.particles
        remote_particles.add_particles(local_particles)

        colliding_particles = instance.particles.select_colliding()

        self.assertEqual(len(colliding_particles), 2)
        self.assertEqual(colliding_particles.mass[0], 4.0 | units.kg)
        self.assertEqual(colliding_particles.mass[1], 5.0 | units.kg)

        attribute_names = dir(instance)

        self.assertTrue('particles' in attribute_names)
        self.assertTrue('get_colliding_indices' in attribute_names)

    def test3(self):
        original = ParticlesWithBindingInterface()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('METHOD')
        handler.add_method('get_mass', (handler.NO_UNIT,), (units.kg, handler.ERROR_CODE))
        handler.add_method('set_mass', (handler.NO_UNIT, units.kg,), (handler.ERROR_CODE,))
        handler.add_method('get_next', (handler.INDEX,), (handler.LINK("particles"), handler.ERROR_CODE))
        handler.add_method('new_particle', (units.kg,), (handler.NO_UNIT, handler.ERROR_CODE))
        handler.add_method('delete_particle', (handler.NO_UNIT,), (handler.ERROR_CODE,))
        handler.add_method('get_number_of_particles', (), (handler.NO_UNIT, handler.ERROR_CODE,))

        handler = instance.get_handler('PARTICLES')
        handler.define_set('particles', 'id')
        handler.set_new('particles', 'new_particle')
        handler.set_delete('particles', 'delete_particle')
        handler.add_setter('particles', 'set_mass')
        handler.add_getter('particles', 'get_mass', names=('mass',))
        handler.add_getter('particles', 'get_next', names=('next_particle',))

        local_particles = datamodel.Particles(4)
        local_particles.mass = units.kg.new_quantity([3.0, 4.0, 5.0, 6.0])

        remote_particles = instance.particles
        remote_particles.add_particles(local_particles)

        self.assertEqual(len(instance.particles), 4)
        self.assertEqual(instance.particles[0].next_particle, instance.particles[1])
        self.assertEqual(instance.particles[1].next_particle, instance.particles[2])
        self.assertEqual(instance.particles[2].next_particle, instance.particles[3])
        self.assertEqual(instance.particles[3].next_particle, None)

    def test4(self):
        original = ParticlesWithBindingInterface()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('METHOD')
        handler.add_method('get_mass', (handler.NO_UNIT,), (units.kg, handler.ERROR_CODE))
        handler.add_method('set_mass', (handler.NO_UNIT, units.kg,), (handler.ERROR_CODE,))
        handler.add_method('new_particle', (units.kg,), (handler.NO_UNIT, handler.ERROR_CODE))
        handler.add_method('delete_particle', (handler.NO_UNIT,), (handler.ERROR_CODE,))
        handler.add_method('get_number_of_particles', (), (handler.NO_UNIT, handler.ERROR_CODE,))

        handler.add_method('add_1_to_mass', (handler.INDEX,), (handler.ERROR_CODE,))

        handler = instance.get_handler('PARTICLES')
        handler.define_set('particles', 'id')
        handler.set_new('particles', 'new_particle')
        handler.set_delete('particles', 'delete_particle')
        handler.add_setter('particles', 'set_mass')
        handler.add_getter('particles', 'get_mass', names=('mass',))
        handler.add_method('particles', 'add_1_to_mass', 'add_one')

        local_particles = datamodel.Particles(4)
        local_particles.mass = units.kg.new_quantity([3.0, 4.0, 5.0, 6.0])

        remote_particles = instance.particles
        remote_particles.add_particles(local_particles)

        self.assertEqual(len(instance.particles), 4)
        self.assertEqual(instance.particles.mass, units.kg.new_quantity([3.0, 4.0, 5.0, 6.0]))
        instance.particles[0].add_one()
        self.assertEqual(instance.particles.mass, units.kg.new_quantity([4.0, 4.0, 5.0, 6.0]))
        instance.particles.add_one()
        self.assertEqual(instance.particles.mass, units.kg.new_quantity([5.0, 5.0, 6.0, 7.0]))

    def test5(self):
        original = ParticlesWithBindingInterface()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('METHOD')
        handler.add_method('get_mass', (handler.NO_UNIT,), (nbody_system.mass, handler.ERROR_CODE))
        handler.add_method('set_mass', (handler.NO_UNIT, nbody_system.mass,), (handler.ERROR_CODE,))
        handler.add_method('new_particle', (nbody_system.mass,), (handler.NO_UNIT, handler.ERROR_CODE))
        handler.add_method('delete_particle', (handler.NO_UNIT,), (handler.ERROR_CODE,))
        handler.add_method('get_number_of_particles', (), (handler.NO_UNIT, handler.ERROR_CODE,))

        handler = instance.get_handler('PARTICLES')
        handler.define_set('particles', 'id')
        handler.set_new('particles', 'new_particle')
        handler.set_delete('particles', 'delete_particle')
        handler.add_setter('particles', 'set_mass')
        handler.add_getter('particles', 'get_mass', names=('mass',))

        convert_nbody = nbody_system.nbody_to_si(10.0 | units.kg, 5.0 | units.m)

        handler = instance.get_handler('UNIT')
        handler.set_nbody_converter(convert_nbody)

        local_particles = datamodel.Particles(4)
        local_particles.mass = units.kg.new_quantity([3.0, 4.0, 5.0, 6.0])

        remote_particles = instance.particles
        remote_particles.add_particles(local_particles)

        self.assertEqual(len(original.masses), 4)
        self.assertAlmostEqual(original.masses[0], 0.3, 5)
        self.assertAlmostEqual(original.masses[3], 0.6, 5)

        self.assertEqual(len(instance.particles), 4)
        self.assertEqual(instance.particles[0].mass, 3.0 | units.kg)

    def test6(self):
        original = ParticlesWithBindingInterface()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('METHOD')
        handler.add_method('get_mass', (handler.NO_UNIT,), (nbody_system.mass, handler.ERROR_CODE))
        handler.add_method('set_mass', (handler.NO_UNIT, nbody_system.mass,), (handler.ERROR_CODE,))
        handler.add_method('new_particle', (nbody_system.mass,), (handler.NO_UNIT, handler.ERROR_CODE))
        handler.add_method('delete_particle', (handler.NO_UNIT,), (handler.ERROR_CODE,))
        handler.add_method('get_number_of_particles', (), (handler.NO_UNIT, handler.ERROR_CODE,))

        handler = instance.get_handler('PARTICLES')
        handler.define_set('particles', 'id')
        handler.set_new('particles', 'new_particle')
        handler.set_delete('particles', 'delete_particle')
        handler.add_setter('particles', 'set_mass')
        handler.add_getter('particles', 'get_mass', names=('mass',))
        handler.add_subselect_in_set('particles', 'get_heaviest_particle', public_name='heaviest')
        handler.add_subselect_in_set('particles', 'get_lightest_particles', get_number_of_particles_name='get_number_of_lightest_particles', public_name='lightest')

        convert_nbody = nbody_system.nbody_to_si(10.0 | units.kg, 5.0 | units.m)

        handler = instance.get_handler('UNIT')
        handler.set_nbody_converter(convert_nbody)

        local_particles = datamodel.Particles(4)
        local_particles.mass = units.kg.new_quantity([3.0, 4.0, 6.0, 5.0])

        remote_particles = instance.particles
        remote_particles.add_particles(local_particles)

        self.assertEqual(len(instance.particles), 4)
        self.assertEqual(instance.particles[0].mass, 3.0 | units.kg)
        subset = instance.particles.heaviest()
        self.assertEqual(len(subset), 1)
        self.assertEqual(subset[0], local_particles[2])

        subset = instance.particles.lightest()
        self.assertEqual(len(subset), 2)
        self.assertEqual(subset[0], local_particles[0])
        self.assertEqual(subset[1], local_particles[1])

    def test7(self):
        original = ParticlesWithBindingInterface()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('METHOD')
        handler.add_method('get_mass', (handler.NO_UNIT,), (nbody_system.mass, handler.ERROR_CODE))
        handler.add_method('set_mass', (handler.NO_UNIT, nbody_system.mass,), (handler.ERROR_CODE,))
        handler.add_method('new_particle', (nbody_system.mass,), (handler.NO_UNIT, handler.ERROR_CODE))
        handler.add_method('delete_particle', (handler.NO_UNIT,), (handler.ERROR_CODE,))
        handler.add_method('get_number_of_particles', (), (handler.NO_UNIT, handler.ERROR_CODE,))

        handler = instance.get_handler('PARTICLES')
        handler.define_set('particles', 'id')
        handler.set_new('particles', 'new_particle')
        handler.set_delete('particles', 'delete_particle')
        handler.add_setter('particles', 'set_mass')
        handler.add_getter('particles', 'get_mass', names=('mass',))
        handler.add_subselect_from_particle('particles', 'get_particle_with_same_mass', get_number_of_particles_name='get_number_of_particles_with_same_mass', public_name='samemass')

        convert_nbody = nbody_system.nbody_to_si(1.0 | units.kg, 5.0 | units.m)

        handler = instance.get_handler('UNIT')
        handler.set_nbody_converter(convert_nbody)

        local_particles = datamodel.Particles(4)
        local_particles.mass = units.kg.new_quantity([3.5, 4.0, 6.0, 4.5])

        remote_particles = instance.particles
        remote_particles.add_particles(local_particles)

        self.assertEqual(len(instance.particles), 4)
        self.assertEqual(instance.particles[1].mass, 4.0 | units.kg)
        subset = instance.particles[1].samemass()
        self.assertEqual(len(subset), 2)
        self.assertEqual(subset[0], local_particles[0])
        self.assertEqual(subset[1], local_particles[3])

    def test8(self):
        original = ParticlesWithBindingInterface()
        original.masses[1] = 10
        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('METHOD')
        handler.add_method('get_mass', (handler.NO_UNIT,), (units.kg, handler.ERROR_CODE))
        handler.add_method('get_next', (handler.INDEX,), (handler.LINK("particles"), handler.ERROR_CODE))
        pickled_handler = pickle.dumps(handler)
        unpickled_handler = pickle.loads(pickled_handler)
        method1 = handler.get_attribute('get_mass', original.get_mass)
        method2 = unpickled_handler.get_attribute('get_mass', unpickled_handler.interface.legacy_interface.get_mass)
        self.assertTrue(method1 is not None)
        self.assertTrue(method2 is not None)
        self.assertAlmostRelativeEquals(method1([1])[0], 10 | units.kg)
        self.assertAlmostRelativeEquals(method2([1])[0], 10 | units.kg)
        original.masses[1] = 20
        self.assertAlmostRelativeEquals(method1([1]), 20 | units.kg)
        self.assertAlmostRelativeEquals(method2([1]), 10 | units.kg)

    def test9(self):
        original = ParticlesWithBindingInterface()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('METHOD')
        handler.add_method('get_mass', (handler.NO_UNIT,), (units.kg, handler.ERROR_CODE))
        handler.add_method('set_mass', (handler.NO_UNIT, units.kg,), (handler.ERROR_CODE,))
        handler.add_method('new_particle', (units.kg,), (handler.NO_UNIT, handler.ERROR_CODE))
        handler.add_method('delete_particle', (handler.NO_UNIT,), (handler.ERROR_CODE,))
        handler.add_method('get_number_of_particles', (), (handler.NO_UNIT, handler.ERROR_CODE,))

        handler = instance.get_handler('PARTICLES')
        handler.define_set('particles', 'id')
        handler.set_new('particles', 'new_particle')
        handler.set_delete('particles', 'delete_particle')
        handler.add_setter('particles', 'set_mass')
        handler.add_getter('particles', 'get_mass', names=('mass',))

        local_particles = datamodel.Particles(4)
        local_particles.mass = units.kg.new_quantity([3.0, 4.0, 5.0, 6.0])

        remote_particles = instance.particles
        remote_particles.add_particles(local_particles)

        self.assertEqual(len(original.masses), 4)
        self.assertEqual(original.masses[0], 3.0)
        self.assertEqual(original.masses[3], 6.0)

        pickled_instance = pickle.dumps(instance)
        unpickled_instance = pickle.loads(pickled_instance)

        self.assertEqual(len(unpickled_instance.particles), 4)
        self.assertEqual(unpickled_instance.particles[1].mass, 4.0 | units.kg)


class TestGridWithBinding(amusetest.TestCase):
    class TestInterface(object):

        shape = (11, 5, 5)

        def __init__(self):
            self.storage = numpy.arange(
                self.shape[0]*self.shape[1]*self.shape[2]).reshape(self.shape)

        def get_range(self):
            return (0, self.shape[0]-1, 0, self.shape[1]-1, 0, self.shape[2]-1)

        def get_position(self, i_s, j_s, k_s):
            return [numpy.asarray(i_s), numpy.asarray(j_s)]

        def get_a(self, i_s, j_s, k_s):
            return [numpy.asarray([(self.storage[i][j][k]) for i, j, k in zip(i_s, j_s, k_s)]),]

        def set_a(self, i_s, j_s, k_s, values):
            index = 0
            for i, j, k in zip(i_s, j_s, k_s):
                self.storage[i][j][k] = values[index].value_in(units.m)
                index += 1

    def test1(self):
        original = self.TestInterface()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('METHOD')
        handler.add_method('get_a', (handler.INDEX, handler.INDEX, handler.INDEX,), (units.kg,))
        handler.add_method('set_a', (handler.INDEX, handler.INDEX, handler.INDEX, units.kg,), ())

        self.assertEqual(instance.get_a([1], [2], [3]), [38] | units.kg)

        handler = instance.get_handler('PARTICLES')
        handler.define_grid('grid',)
        handler.add_setter('grid', 'set_a', names=('mass',))
        handler.add_getter('grid', 'get_a', names=('mass',))

        grid = instance.grid

        self.assertEqual(grid[1][2][3].mass, 38 | units.kg)
        self.assertEqual(grid[0:2][1][2][3].mass, 38 | units.kg)
        self.assertEqual(len(grid[1][2].mass), 5)
        self.assertTrue(numpy.all(grid[1][2].mass == [35, 36, 37, 38, 39] | units.kg))

    def test2(self):
        original = self.TestInterface()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('METHOD')
        handler.add_method('get_position', (handler.INDEX, handler.INDEX, handler.INDEX,), (units.m, units.m,))

        handler = instance.get_handler('PARTICLES')
        handler.define_grid('grid', axes_names=['x', 'y'])
        handler.add_getter('grid', 'get_position', names=('x', 'y',))

        grid = instance.grid

        self.assertEqual(grid[1][2][3].position, [1, 2] | units.m)
        self.assertEqual(grid[1][2][1].position, [1, 2] | units.m)
        self.assertEqual(grid[1][2][2].position, [1, 2] | units.m)
        self.assertEqual(grid[1][2][0].position, [1, 2] | units.m)
        self.assertEqual(grid[1][2].position, [[1, 2], [1, 2], [1, 2], [1, 2], [1, 2]] | units.m)
        self.assertEqual(grid[0][1][1].position, [0, 1] | units.m)

    def test3(self):
        original = self.TestInterface()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('METHOD')
        handler.add_method('get_position', (handler.INDEX, handler.INDEX, handler.INDEX,), (units.m, units.m,))

        handler = instance.get_handler('PARTICLES')
        handler.define_grid('grid', axes_names=['x', 'y'])
        handler.add_getter('grid', 'get_position', names=('x', 'y',))

        grid = instance.grid

        self.assertEqual(grid.__class__.__name__, "Grid")
        self.assertTrue(isinstance(grid, datamodel.RegularGrid))

    def test4(self):
        original = self.TestInterface()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('METHOD')
        handler.add_method('get_position', (handler.INDEX, handler.INDEX, handler.INDEX,), (units.m, units.m,))

        handler = instance.get_handler('PARTICLES')
        handler.define_grid('grid', axes_names=['x', 'y'], grid_class=datamodel.CartesianGrid)
        handler.add_getter('grid', 'get_position', names=('x', 'y',))

        grid = instance.grid

        self.assertEqual(grid.__class__.__name__, "CartesianGrid")


class TestGridWithBinding2(amusetest.TestCase):
    class TestInterface(object):

        shape = (11, 5)

        def __init__(self):
            self.storage = numpy.arange(
                self.shape[0]*self.shape[1]).reshape(self.shape)

        def get_range(self):
            return (0, self.shape[0]-1, 0, self.shape[1]-1)

        def get_position(self, i_s):
            return [numpy.asarray(i_s)]

        def get_a(self, i_s, j_s):
            # print "indices:", i_s, j_s, k_s
            # print "values:", numpy.asarray([(self.storage[i][j][k]) for i,j,k in zip(i_s, j_s, k_s)])
            return [numpy.asarray([(self.storage[i][j]) for i, j in zip(i_s, j_s)]),]

        def set_a(self, i_s, j_s, values):
            index = 0
            for i, j in zip(i_s, j_s):
                self.storage[i][j] = values[index]
                index += 1

    def test1(self):
        original = self.TestInterface()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('METHOD')
        handler.add_method('get_a', (handler.INDEX, handler.INDEX,), (units.kg,))
        handler.add_method('set_a', (handler.INDEX, handler.INDEX, units.kg,), ())

        self.assertEqual(instance.get_a([1], [2]), [7] | units.kg)

        handler = instance.get_handler('PARTICLES')
        handler.define_grid('grid',)
        handler.add_setter('grid', 'set_a', names=('mass',))
        handler.add_getter('grid', 'get_a', names=('mass',))

        grid = instance.grid

        self.assertEqual(grid[1, 2].mass, original.storage[1, 2] | units.kg)
        self.assertEqual(len(grid[1].mass), len(original.storage[1]))

        grid[1, 1:].mass = [5. | units.kg]*4

        self.assertEqual(original.storage[1, 1:], 5)
        self.assertEqual(grid[1, 1:].mass, original.storage[1, 1:] | units.kg)


class TestGridWithBinding3(amusetest.TestCase):
    class TestInterface(object):

        shape = (11, 5)

        def __init__(self):
            self.storage = numpy.arange(
                self.shape[0]*self.shape[1]).reshape(self.shape)

        def get_range(self):
            return (0, self.shape[0]-1)

        def get_range2(self):
            return (0, self.shape[1]-1)

        def get_a(self, i_s, j_s):
            return [numpy.asarray([(self.storage[i][j]) for i, j in zip(i_s, j_s)]),]

        def set_a(self, i_s, j_s, values):
            for i, j, val in zip(i_s, j_s, values):
                self.storage[i][j] = val

    def test1(self):
        original = self.TestInterface()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('METHOD')
        handler.add_method('get_a', (handler.INDEX, handler.INDEX,), (units.kg,))
        handler.add_method('set_a', (handler.INDEX, handler.INDEX, units.kg,), ())

        self.assertEqual(instance.get_a([1], [2]), [7] | units.kg)

        handler = instance.get_handler('PARTICLES')
        handler.define_grid('grid',)
        handler.add_gridded_setter('grid', 'set_a', 'get_range2', names=('mass',))
        handler.add_gridded_getter('grid', 'get_a', 'get_range2', names=('mass',))

        grid = instance.grid

        self.assertEqual(grid[1].mass, original.storage[1] | units.kg)
        self.assertEqual(len(grid[1].mass), len(original.storage[1]))

        self.assertEqual(grid[2:4].mass, original.storage[2:4] | units.kg)

        self.assertEqual(grid[-3:].mass, original.storage[-3:] | units.kg)

        grid[1].mass = [5.]*5 | units.kg

        self.assertEqual(original.storage[1], 5)
        self.assertEqual(grid[1].mass, original.storage[1] | units.kg)


class TestGridWithBinding4(amusetest.TestCase):
    class TestInterface(object):

        shape = (3, 4, 5)

        def __init__(self):
            self.storage = numpy.arange(
                self.shape[0]*self.shape[1]*self.shape[2]).reshape(self.shape)

        def get_range(self):
            return (0, self.shape[0]-1, 0, self.shape[1]-1)

        def get_range2(self):
            return (0, self.shape[2]-1)

        def get_a(self, i_s, j_s, k_s):
            return [numpy.asarray([self.storage[i, j, k] for i, j, k in zip(i_s, j_s, k_s)]),]

        def set_a(self, i_s, j_s, k_s, values):
            for i, j, k, val in zip(i_s, j_s, k_s, values):
                self.storage[i, j, k] = val

    def test1(self):
        original = self.TestInterface()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('METHOD')
        handler.add_method('get_a', (handler.INDEX, handler.INDEX, handler.INDEX,), (units.kg,))
        handler.add_method('set_a', (handler.INDEX, handler.INDEX, handler.INDEX, units.kg,), ())

        self.assertEqual(instance.get_a([1], [2], [3]), [33] | units.kg)

        handler = instance.get_handler('PARTICLES')
        handler.define_grid('grid',)
        handler.add_gridded_setter('grid', 'set_a', 'get_range2', names=('mass',))
        handler.add_gridded_getter('grid', 'get_a', 'get_range2', names=('mass',))

        grid = instance.grid

        self.assertEqual(grid[1, 1].mass, original.storage[1, 1] | units.kg)
        self.assertEqual(grid[0:2, 1:3].mass, original.storage[0:2, 1:3] | units.kg)
        self.assertEqual(len(grid[1].mass), len(original.storage[1]))
        self.assertEqual(grid[2:4].mass, original.storage[2:4] | units.kg)
        self.assertEqual(grid[-3:].mass, original.storage[-3:] | units.kg)

        grid[1].mass = numpy.ones((4, 5))*5 | units.kg

        self.assertEqual(original.storage[1], 5)
        self.assertEqual(grid[1].mass, original.storage[1] | units.kg)


class TestGridWithBinding5(amusetest.TestCase):
    class TestInterface(object):

        shape = (3, 4, 5, 6)

        def __init__(self):
            self.storage = numpy.arange(
                numpy.prod(self.shape)).reshape(self.shape)

        def get_range(self):
            return (0, self.shape[0]-1, 0, self.shape[1]-1)

        def get_range2(self):
            return (0, self.shape[2]-1, 0, self.shape[3]-1)

        def get_a(self, i_s, j_s, k_s, l_s):
            return [numpy.asarray([self.storage[i, j, k, l] for i, j, k, l in zip(i_s, j_s, k_s, l_s)]),]

        def set_a(self, i_s, j_s, k_s, l_s, values):
            for i, j, k, l, val in zip(i_s, j_s, k_s, l_s, values):
                self.storage[i, j, k, l] = val

    def test1(self):
        original = self.TestInterface()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('METHOD')
        handler.add_method('get_a', (handler.INDEX, handler.INDEX, handler.INDEX, handler.INDEX,), (units.kg,))
        handler.add_method('set_a', (handler.INDEX, handler.INDEX, handler.INDEX, handler.INDEX, units.kg,), ())

        self.assertEqual(instance.get_a([1], [2], [3], [4]), original.storage[1, 2, 3, 4] | units.kg)

        handler = instance.get_handler('PARTICLES')
        handler.define_grid('grid',)
        handler.add_gridded_setter('grid', 'set_a', 'get_range2', names=('mass',))
        handler.add_gridded_getter('grid', 'get_a', 'get_range2', names=('mass',))

        grid = instance.grid

        self.assertEqual(grid[1, 1].mass, original.storage[1, 1] | units.kg)
        self.assertEqual(grid[0:2, 1:3].mass, original.storage[0:2, 1:3] | units.kg)
        self.assertEqual(len(grid[1].mass), len(original.storage[1]))
        self.assertEqual(grid[2:4].mass, original.storage[2:4] | units.kg)
        self.assertEqual(grid[-3:].mass, original.storage[-3:] | units.kg)

        grid[1].mass = numpy.ones((4, 5, 6))*5 | units.kg

        self.assertEqual(original.storage[1], 5)
        self.assertEqual(grid[1].mass, original.storage[1] | units.kg)

# attempt to test string returns, needs improvement


class TestGridWithBinding6(amusetest.TestCase):
    class TestInterface(object):

        shape = (3, 4, 5, 6)

        def __init__(self):
            self.storage = numpy.arange(
                numpy.prod(self.shape)).reshape(self.shape).astype(numpy.dtype(str))

        def get_range(self):
            return (0, self.shape[0]-1, 0, self.shape[1]-1)

        def get_range2(self):
            return (0, self.shape[2]-1, 0, self.shape[3]-1)

        def get_a(self, i_s, j_s, k_s, l_s):
            return [numpy.asarray([self.storage[i, j, k, l] for i, j, k, l in zip(i_s, j_s, k_s, l_s)]),]

        def set_a(self, i_s, j_s, k_s, l_s, values):
            for i, j, k, l, val in zip(i_s, j_s, k_s, l_s, values):
                self.storage[i, j, k, l] = val

    def test1(self):
        original = self.TestInterface()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('METHOD')
        handler.add_method('get_a', (handler.INDEX, handler.INDEX, handler.INDEX, handler.INDEX,), (handler.NO_UNIT,))
        handler.add_method('set_a', (handler.INDEX, handler.INDEX, handler.INDEX, handler.INDEX, handler.NO_UNIT,), ())

        self.assertEqual(instance.get_a([1], [2], [3], [4]), original.storage[1, 2, 3, 4])

        handler = instance.get_handler('PARTICLES')
        handler.define_grid('grid',)
        handler.add_gridded_setter('grid', 'set_a', 'get_range2', names=('mass',))
        handler.add_gridded_getter('grid', 'get_a', 'get_range2', names=('mass',))

        grid = instance.grid

        self.assertEqual(grid[1, 1].mass, original.storage[1, 1])
        self.assertEqual(grid[0:2, 1:3].mass, original.storage[0:2, 1:3])
        self.assertEqual(len(grid[1].mass), len(original.storage[1]))
        self.assertEqual(grid[2:4].mass, original.storage[2:4])
        self.assertEqual(grid[-3:].mass, original.storage[-3:])

        grid[1].mass = numpy.ones((4, 5, 6))*5

#        self.assertEquals(original.storage[1],'5')
        self.assertEqual(grid[1].mass, original.storage[1])


class TestGridWithBinding7(amusetest.TestCase):
    class TestInterface(object):

        shape = ()

        def __init__(self):
            self.storage = 123.

        def get_range(self):
            return ()

        def get_a(self):
            return self.storage

        def set_a(self, value):
            self.storage = value

    def test1(self):
        original = self.TestInterface()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('METHOD')
        handler.add_method('get_a', (), (units.kg,))
        handler.add_method('set_a', (units.kg,), ())

        handler = instance.get_handler('PARTICLES')
        handler.define_grid('grid',)
        handler.add_setter('grid', 'set_a', names=('mass',))
        handler.add_getter('grid', 'get_a', names=('mass',))

        grid = instance.grid
        self.assertEqual(grid.mass, 123 | units.kg)

    def test2(self):
        original = self.TestInterface()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('METHOD')
        handler.add_method('get_a', (), (units.kg,))
        handler.add_method('set_a', (units.kg,), ())

        handler = instance.get_handler('PARTICLES')
        handler.define_grid('grid',)
        handler.add_setter('grid', 'set_a', names=('mass',))
        handler.add_getter('grid', 'get_a', names=('mass',))

        grid = instance.grid

        grid.mass = 321 | units.kg
        self.assertEqual(original.storage, 321)
        self.assertEqual(grid.mass, 321 | units.kg)


class TestGridWithBinding8(amusetest.TestCase):
    class TestInterface(object):

        shape = ()

        def __init__(self):
            self.storage1 = 12.
            self.storage2 = 123.

        def get_range(self):
            return ()

        def get_a(self):
            return self.storage1, self.storage2

        def set_a(self, value1, value2):
            self.storage1 = value1
            self.storage2 = value2

    def test1(self):
        original = self.TestInterface()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('METHOD')
        handler.add_method('get_a', (), (units.kg, units.m))
        handler.add_method('set_a', (units.kg, units.m), ())

        handler = instance.get_handler('PARTICLES')
        handler.define_grid('grid',)
        handler.add_setter('grid', 'set_a', names=('mass', 'l'))
        handler.add_getter('grid', 'get_a', names=('mass', 'l'))

        grid = instance.grid
        self.assertEqual(grid.mass, 12 | units.kg)
        self.assertEqual(grid.l, 123 | units.m)

    def test2(self):
        original = self.TestInterface()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('METHOD')
        handler.add_method('get_a', (), (units.kg, units.m))
        handler.add_method('set_a', (units.kg, units.m), ())

        handler = instance.get_handler('PARTICLES')
        handler.define_grid('grid',)
        handler.add_setter('grid', 'set_a', names=('mass', 'l'))
        handler.add_getter('grid', 'get_a', names=('mass', 'l'))

        grid = instance.grid

        grid1 = grid.copy()

        grid1.mass = 321 | units.kg
        grid1.l = 32 | units.m

        grid1.new_channel_to(grid).copy_all_attributes()

        self.assertEqual(original.storage1, 321)
        self.assertEqual(original.storage2, 32)
        self.assertEqual(grid.mass, 321 | units.kg)
        self.assertEqual(grid.l, 32 | units.m)


class CodeInterfaceAndLegacyFunctionsTest(amusetest.TestCase):

    def test1(self):
        class TestClass(object):

            @legacy_function
            def echo_inputs():
                function = LegacyFunctionSpecification()
                function.addParameter('input1', dtype='d', direction=function.OUT)
                function.addParameter('input2', dtype='d', direction=function.OUT)
                function.addParameter('output1', dtype='d', direction=function.IN)
                function.addParameter('output2', dtype='d', direction=function.IN)
                function.result_type = 'i'
                return function

        original = TestClass()
        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('METHOD')
        handler.add_method('echo_inputs', (units.m, units.s), (units.s, handler.ERROR_CODE))
        self.assertRaises(exceptions.AmuseException, lambda: instance.echo_inputs,
            expected_message="Incorrect definition of method 'echo_inputs' of class 'InCodeComponentImplementation', "
            "the number of outputs do not match, expected 3, actual 2.")

        instance = interface.InCodeComponentImplementation(original)
        handler = instance.get_handler('METHOD')
        handler.add_method('echo_inputs', (units.m, units.s), (units.s, units.m, units.s, handler.ERROR_CODE))
        self.assertRaises(exceptions.AmuseException, lambda: instance.echo_inputs,
            expected_message="Incorrect definition of method 'echo_inputs' of class 'InCodeComponentImplementation', "
            "the number of outputs do not match, expected 3, actual 4.")

    def test2(self):
        class TestClass(object):

            @legacy_function
            def echo_inputs():
                function = LegacyFunctionSpecification()
                function.addParameter('input1', dtype='d', direction=function.OUT)
                function.addParameter('input2', dtype='d', direction=function.OUT)
                function.addParameter('output1', dtype='d', direction=function.IN)
                function.addParameter('output2', dtype='d', direction=function.IN)
                function.result_type = 'i'
                return function

        original = TestClass()

        instance = interface.InCodeComponentImplementation(original)

        handler = instance.get_handler('METHOD')
        handler.add_method('echo_inputs', (units.m), (units.s, units.m, handler.ERROR_CODE), )
        self.assertRaises(exceptions.AmuseException, lambda: instance.echo_inputs,
            expected_message="Incorrect definition of method 'echo_inputs' of class 'InCodeComponentImplementation', "
            "the number of inputs do not match, expected 2, actual 1.")

    def test3(self):
        class TestClass(object):
            async_request = None

            @legacy_function
            def echo_inputs():
                function = LegacyFunctionSpecification()
                function.addParameter('input1', dtype='d', direction=function.OUT)
                function.addParameter('input2', dtype='d', direction=function.OUT)
                function.addParameter('output1', dtype='d', direction=function.IN)
                function.addParameter('output2', dtype='d', direction=function.IN)
                function.result_type = 'i'
                return function

        original = TestClass()
        self.assertRaises(exceptions.CodeException, original.echo_inputs, 1, 2,
            expected_message="Exception when calling function 'echo_inputs', of code 'TestClass', exception was ''TestClass' object has no attribute 'channel''")



from amuse.test import amusetest
import numpy


class harmonic_oscillator(object):
    def __init__(self, x, v, method=None):
        self.x = x
        self.v = v
        self.model_time = 0
        self.method = method

    def kick(self, dt):
        self.v += -self.x*dt

    def drift(self, dt):
        self.x += self.v*dt

    def total_energy(self):
        return (self.v**2+self.x**2)/2

    def evolve_step(self, dt):
        self.method(self.kick, self.drift, dt)
        self.model_time += dt


def run_harmonic_oscillator(dt, method):
    h = harmonic_oscillator(0., 1., method=method)
    tend = 100*2*numpy.pi

    data = dict()
    data['x'] = []
    data['time'] = []
    data['de'] = []
    E0 = h.total_energy()

    while h.model_time < tend-dt/2:
        h.evolve_step(dt)
        data['x'].append(h.x)
        data['time'].append(h.model_time)
        E = h.total_energy()
        data['de'].append(abs(E0-E)/E0)
    return data


class TestSymplecticCompositions(amusetest.TestCase):
    def test1(self):
        from amuse.ext.composition_methods import LEAPFROG
        dt1 = .1
        data = run_harmonic_oscillator(dt1, method=LEAPFROG)
        de1 = max(data['de'])
        dt2 = 0.01
        data = run_harmonic_oscillator(dt2, method=LEAPFROG)
        de2 = max(data['de'])
        order = int(numpy.log(de2/de1)/numpy.log(dt2/dt1)+0.5)
        self.assertEqual(order, 2)

    def test2(self):
        from amuse.ext.composition_methods import SPLIT_4TH_S_M6
        dt1 = .5
        data = run_harmonic_oscillator(dt1, method=SPLIT_4TH_S_M6)
        de1 = max(data['de'])
        dt2 = 0.05
        data = run_harmonic_oscillator(dt2, method=SPLIT_4TH_S_M6)
        de2 = max(data['de'])
        order = int(numpy.log(de2/de1)/numpy.log(dt2/dt1)+0.5)
        self.assertEqual(order, 4)

    def test3(self):
        from amuse.ext.composition_methods import SPLIT_4TH_S_M5
        dt1 = .5
        data = run_harmonic_oscillator(dt1, method=SPLIT_4TH_S_M5)
        de1 = max(data['de'])
        dt2 = 0.05
        data = run_harmonic_oscillator(dt2, method=SPLIT_4TH_S_M5)
        de2 = max(data['de'])
        order = int(numpy.log(de2/de1)/numpy.log(dt2/dt1)+0.5)
        self.assertEqual(order, 4)

    def test4(self):
        from amuse.ext.composition_methods import SPLIT_4TH_S_M4
        dt1 = .5
        data = run_harmonic_oscillator(dt1, method=SPLIT_4TH_S_M4)
        de1 = max(data['de'])
        dt2 = 0.05
        data = run_harmonic_oscillator(dt2, method=SPLIT_4TH_S_M4)
        de2 = max(data['de'])
        order = int(numpy.log(de2/de1)/numpy.log(dt2/dt1)+0.5)
        self.assertEqual(order, 4)

    def test5(self):
        from amuse.ext.composition_methods import SPLIT_6TH_SS_M11
        dt1 = .5
        data = run_harmonic_oscillator(dt1, method=SPLIT_6TH_SS_M11)
        de1 = max(data['de'])
        dt2 = 0.05
        data = run_harmonic_oscillator(dt2, method=SPLIT_6TH_SS_M11)
        de2 = max(data['de'])
        order = int(numpy.log(de2/de1)/numpy.log(dt2/dt1)+0.5)
        self.assertEqual(order, 6)

    def test6(self):
        from amuse.ext.composition_methods import SPLIT_6TH_SS_M13
        dt1 = .5
        data = run_harmonic_oscillator(dt1, method=SPLIT_6TH_SS_M13)
        de1 = max(data['de'])
        dt2 = 0.05
        data = run_harmonic_oscillator(dt2, method=SPLIT_6TH_SS_M13)
        de2 = max(data['de'])
        order = int(numpy.log(de2/de1)/numpy.log(dt2/dt1)+0.5)
        self.assertEqual(order, 6)

    def test7(self):
        from amuse.ext.composition_methods import SPLIT_8TH_SS_M21
        dt1 = 1.
        data = run_harmonic_oscillator(dt1, method=SPLIT_8TH_SS_M21)
        de1 = max(data['de'])
        dt2 = 0.25
        data = run_harmonic_oscillator(dt2, method=SPLIT_8TH_SS_M21)
        de2 = max(data['de'])
        order = int(numpy.log(de2/de1)/numpy.log(dt2/dt1)+0.5)
        self.assertEqual(order, 8)

    def test8(self):
        from amuse.ext.composition_methods import SPLIT_10TH_SS_M35
        dt1 = 1.
        data = run_harmonic_oscillator(dt1, method=SPLIT_10TH_SS_M35)
        de1 = max(data['de'])
        dt2 = 0.5
        data = run_harmonic_oscillator(dt2, method=SPLIT_10TH_SS_M35)
        de2 = max(data['de'])
        order = int(numpy.log(de2/de1)/numpy.log(dt2/dt1)+0.5)
        self.assertEqual(order, 10)




# -*- coding: utf-8 -*-
from amuse.test.amusetest import TestWithMPI
import os
import sys
import numpy
import time
import math

from amuse.community.bhtree.interface import BHTreeInterface, BHTree
from amuse.support.exceptions import AmuseException
from amuse.units import constants
from amuse.units import nbody_system
from amuse.units import units
from amuse.units import quantities
from amuse import datamodel
from amuse.datamodel import particle_attributes
from amuse.ic import plummer
try:
    from matplotlib import pyplot
    HAS_MATPLOTLIB = True
except ImportError:
    HAS_MATPLOTLIB = False


class TestBHTreeInterface(TestWithMPI):

    def test0(self):
        instance = BHTreeInterface()
        self.assertTrue("Barnes" in instance.all_literature_references_string())
        instance.stop()

    def test1(self):
        instance = BHTreeInterface()
        instance.initialize_code()
        instance.commit_parameters()
        res1 = instance.new_particle(mass=11.0, radius=2.0, x=0.0, y=0.0, z=0.0, vx=0.0, vy=0.0, vz=0.0)
        res2 = instance.new_particle(mass=21.0, radius=5.0, x=10.0, y=0.0, z=0.0, vx=10.0, vy=0.0, vz=0.0)

        instance.commit_particles()
        self.assertEqual(1, res1['index_of_the_particle'])
        self.assertEqual(2, res2['index_of_the_particle'])

        retrieved_state1 = instance.get_state(1)
        retrieved_state2 = instance.get_state(2)

        self.assertEqual(11.0,  retrieved_state1['mass'])
        self.assertEqual(21.0,  retrieved_state2['mass'])
        self.assertEqual(0.0,  retrieved_state1['x'])
        self.assertEqual(10.0,  retrieved_state2['x'])

        self.assertEqual(1, instance.get_index_of_first_particle()['index_of_the_particle'])
        self.assertEqual(2, instance.get_index_of_next_particle(1)['index_of_the_next_particle'])

        instance.cleanup_code()
        instance.stop()

    def test2(self):
        instance = BHTreeInterface()
        instance.initialize_code()

        instance.commit_parameters()
        for i in [1, 2, 3]:
            temp_particle = instance.new_particle(mass=i, radius=1.0, x=0.0, y=0.0, z=0.0, vx=0.0, vy=0.0, vz=0.0)
            self.assertEqual(i, temp_particle['index_of_the_particle'])

        instance.commit_particles()
        self.assertEqual(1, instance.get_index_of_first_particle()['index_of_the_particle'])
        self.assertEqual(2, instance.get_index_of_next_particle(1)['index_of_the_next_particle'])
        self.assertEqual(3, instance.get_index_of_next_particle(2)['index_of_the_next_particle'])

        instance.delete_particle(1)

        self.assertEqual(2, instance.get_number_of_particles()['number_of_particles'])

        # the deletion does a swap, so 3 is copied to 1, (overwriting old 1 and treesize -> treesize-1
        self.assertEqual(3, instance.get_index_of_first_particle()['index_of_the_particle'])

        self.assertEqual(1, instance.get_index_of_next_particle(2)['__result'])

        instance.cleanup_code()
        instance.stop()

    def test5(self):
        interface = BHTreeInterface()
        interface.initialize_code()

        interface.commit_parameters()
        interface.new_particle([10, 20], [0, 0], [0, 0], [0, 0], [0, 0], [0, 0], [0, 0], [1, 1])
        interface.commit_particles()
        retrieved_state = interface.get_state(1)

        self.assertEqual(10.0,  retrieved_state['mass'])
        self.assertEqual(1, retrieved_state['radius'])

        retrieved_state = interface.get_state([1, 2])
        self.assertEqual(20.0,  retrieved_state['mass'][1])
        self.assertEqual(interface.get_number_of_particles()['number_of_particles'], 2)
        interface.cleanup_code()
        interface.stop()

    def test6(self):
        instance = BHTreeInterface()
        instance.initialize_code()
        instance.commit_parameters()

        ids = []
        for i in [1, 2, 3]:
            id, error = instance.new_particle(mass=i, radius=1.0, x=0.0, y=0.0, z=0.0, vx=0.0, vy=0.0, vz=0.0)
            ids.append(id)

        print(ids)

        instance.commit_particles()

        instance.delete_particle(ids[0])
        id, error = instance.new_particle(mass=4, radius=1.0, x=0.0, y=0.0, z=0.0, vx=0.0, vy=0.0, vz=0.0)
        self.assertNotEqual(id, ids[-1])

        instance.cleanup_code()
        instance.stop()

    def test7(self):
        interface = BHTreeInterface()  # channel_type="remote") #, debugger="xterm")
        interface.initialize_code()

        interface.commit_parameters()
        interface.new_particle([10, 20], [0, 0], [0, 0], [0, 0], [0, 0], [0, 0], [0, 0], [1, 1])
        interface.commit_particles()
        retrieved_state = interface.get_state(1)

        self.assertEqual(10.0,  retrieved_state['mass'])
        self.assertEqual(1, retrieved_state['radius'])

        retrieved_state = interface.get_state([1, 2])
        self.assertEqual(20.0,  retrieved_state['mass'][1])
        self.assertEqual(interface.get_number_of_particles()['number_of_particles'], 2)
        interface.cleanup_code()
        interface.stop()

    def test8(self):
        instance = BHTreeInterface()
        instance.initialize_code()
        instance.set_epsilon_squared(0.1 * 0.1)
        instance.commit_parameters()
        id1, errorcode = instance.new_particle(mass=10.0, radius=1.0, x=0.0, y=0.0, z=0.0, vx=0.0, vy=0.0, vz=0.0)
        id2, errorcode = instance.new_particle(mass=10.0, radius=1.0, x=2.0, y=0.0, z=0.0, vx=10.0, vy=0.0, vz=0.0)

        instance.commit_particles()
        potential, errorcode = instance.get_potential(id1)
        self.assertEqual(errorcode, 0)
        self.assertAlmostRelativeEquals(potential,  -10.0 / numpy.sqrt(2.0**2 + 0.1**2), 8)
        instance.cleanup_code()
        instance.stop()


class TestBHTree(TestWithMPI):
    def new_system_of_sun_and_earth(self):
        stars = datamodel.Stars(2)
        sun = stars[0]
        sun.mass = units.MSun(1.0)
        sun.position = units.m(numpy.array((0.0, 0.0, 0.0)))
        sun.velocity = units.ms(numpy.array((0.0, 0.0, 0.0)))
        sun.radius = units.RSun(1.0)

        earth = stars[1]
        earth.mass = units.kg(5.9736e24)
        earth.radius = units.km(6371)
        earth.position = units.km(numpy.array((149.5e6, 0.0, 0.0)))
        earth.velocity = units.ms(numpy.array((0.0, 29800, 0.0)))

        return stars

    def test1(self):
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.MSun, 149.5e6 | units.km)

        instance = BHTree(convert_nbody)
        instance.parameters.epsilon_squared = 0.001 | units.AU**2

        stars = datamodel.Stars(2)

        sun = stars[0]
        sun.mass = units.MSun(1.0)
        sun.position = [0.0, 0.0, 0.0] | units.m
        sun.velocity = [0.0, 0.0, 0.0] | units.ms
        sun.radius = units.RSun(1.0)

        earth = stars[1]
        earth.mass = units.kg(5.9736e24)
        earth.radius = units.km(6371)
        earth.position = [149.5e6, 0.0, 0.0] | units.km
        earth.velocity = [0.0, 29800, 0.0] | units.ms

        # instance.particles.add_particles(stars)
        instance.particles.add_particles(stars)

        postion_at_start = earth.position.value_in(units.AU)[0]

        instance.evolve_model(365.0 | units.day)
        instance.particles.copy_values_of_all_attributes_to(stars)

        postion_after_full_rotation = earth.position.value_in(units.AU)[0]

        self.assertAlmostEqual(postion_at_start, postion_after_full_rotation, 3)

        instance.evolve_model(365.0 + (365.0 / 2) | units.day)

        instance.particles.copy_values_of_all_attributes_to(stars)

        postion_after_half_a_rotation = earth.position.value_in(units.AU)[0]
        self.assertAlmostEqual(-postion_at_start, postion_after_half_a_rotation, 2)

        instance.evolve_model(365.0 + (365.0 / 2) + (365.0 / 4) | units.day)

        instance.particles.copy_values_of_all_attributes_to(stars)

        postion_after_half_a_rotation = earth.position.value_in(units.AU)[1]

        self.assertAlmostEqual(-postion_at_start, postion_after_half_a_rotation, 1)
        instance.cleanup_code()
        instance.stop()

    def test2(self):
        # not completed
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.MSun, 149.5e6 | units.km)

        instance = BHTree(convert_nbody)
        # instance.dt_dia = 1
        instance.parameters.epsilon_squared = 0.001 | units.AU**2
        # instance.timestep = 0.0001
        # instance.use_self_gravity = 0
        instance.commit_parameters()

        stars = datamodel.Stars(2)
        sun = stars[0]
        sun.mass = units.MSun(1.0)
        sun.position = units.m(numpy.array((0.0, 0.0, 0.0)))
        sun.velocity = units.ms(numpy.array((0.0, 0.0, 0.0)))
        sun.radius = units.RSun(1.0)

        earth = stars[1]
        earth.mass = units.kg(5.9736e24)
        earth.radius = units.km(6371)
        earth.position = units.km(numpy.array((149.5e6, 0.0, 0.0)))
        earth.velocity = units.ms(numpy.array((0.0, 29800, 0.0)))

        instance.particles.add_particles(stars)
        instance.commit_particles()
        self.assertAlmostRelativeEquals(sun.radius, instance.particles[0].radius)

        for x in range(1, 2000, 10):
            instance.evolve_model(x | units.day)
            instance.particles.copy_values_of_all_attributes_to(stars)
            stars.savepoint()

        if HAS_MATPLOTLIB:
            figure = pyplot.figure()
            plot = figure.add_subplot(1, 1, 1)

            x_points = earth.get_timeline_of_attribute("x")
            y_points = earth.get_timeline_of_attribute("y")

            x_points_in_AU = [t_x[1].value_in(units.AU) for t_x in x_points]
            y_points_in_AU = [t_x1[1].value_in(units.AU) for t_x1 in y_points]

            plot.scatter(x_points_in_AU, y_points_in_AU, color="b", marker='o')

            plot.set_xlim(-1.5, 1.5)
            plot.set_ylim(-1.5, 1.5)

            test_results_path = self.get_path_to_results()
            output_file = os.path.join(test_results_path, "bhtree-earth-sun.svg")
            figure.savefig(output_file)

        instance.cleanup_code()
        instance.stop()

    def test3(self):
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.MSun, 149.5e6 | units.km)

        instance = BHTree(convert_nbody)
        # instance.dt_dia = 1
        instance.parameters.epsilon_squared = 0.001 | units.AU**2
        # instance.timestep = 0.0001
        # instance.use_self_gravity = 0
        instance.commit_parameters()

        stars = datamodel.Stars(2)
        star1 = stars[0]
        star2 = stars[1]

        star1.mass = units.MSun(1.0)
        star1.position = units.AU(numpy.array((-.10, 0.0, 0.0)))
        star1.velocity = units.AUd(numpy.array((0.0, 0.0, 0.0)))
        star1.radius = units.RSun(1.0)

        star2.mass = units.MSun(1.0)
        star2.position = units.AU(numpy.array((.10, 0.0, 0.0)))
        star2.velocity = units.AUd(numpy.array((0.0, 0.0, 0.0)))
        star2.radius = units.RSun(100.0)

        instance.particles.add_particles(stars)
        instance.commit_particles()

        for x in range(1, 200, 1):
            instance.evolve_model(x | units.day)
            instance.particles.copy_values_of_all_attributes_to(stars)
            # instance.get_indices_of_colliding_particles()
            # print stars[0].position-stars[1].position
            stars.savepoint()

        instance.cleanup_code()
        instance.stop()

    def test4(self):
        convert_nbody = nbody_system.nbody_to_si(5.0 | units.kg, 10.0 | units.m)

        instance = BHTree(convert_nbody)
        instance.commit_parameters()

        index = instance.new_particle(
            15.0 | units.kg,
            10.0 | units.m, 20.0 | units.m, 30.0 | units.m,
            # 1.0 | units.m/units.s, 1.0 | units.m/units.s, 3.0 | units.m/units.s
            0.0 | units.m/units.s, 0.0 | units.m/units.s, 0.0 | units.m/units.s,
            10.0 | units.m
        )
        instance.commit_particles()
        self.assertEqual(instance.get_mass(index), 15.0 | units.kg)
        self.assertEqual(instance.get_radius(index), 10.0 | units.m)
        instance.cleanup_code()
        instance.stop()

    def test5(self):

        instance = BHTree()
        instance.commit_parameters()

        index = instance.new_particle(
            15.0 | nbody_system.mass,
            10.0 | nbody_system.length, 20.0 | nbody_system.length, 30.0 | nbody_system.length,
            1.0 | nbody_system.speed, 1.0 | nbody_system.speed, 3.0 | nbody_system.speed,
            10.0 | nbody_system.length
        )
        instance.commit_particles()
        self.assertEqual(instance.get_mass(index), 15.0 | nbody_system.mass)
        self.assertEqual(instance.get_radius(index), 10.0 | nbody_system.length)

        instance.cleanup_code()
        instance.stop()

    def test6(self):
        convert_nbody = nbody_system.nbody_to_si(5.0 | units.kg, 10.0 | units.m)
        instance = BHTree(convert_nbody)
        instance.commit_parameters()

        indices = instance.new_particle(
            [15.0, 30.0] | units.kg,
            [10.0, 20.0] | units.m, [20.0, 40.0] | units.m, [30.0, 50.0] | units.m,
            # 1.0 | units.m/units.s, 1.0 | units.m/units.s, 3.0 | units.m/units.s
            [0.0, 0.01] | units.m/units.s, [0.0, 0.01] | units.m/units.s, [0.0, 0.01] | units.m/units.s,
            [10.0, 20.0] | units.m
        )
        instance.commit_particles()

        self.assertEqual(instance.get_mass(indices[0]), 15.0 | units.kg)
        self.assertEqual(instance.get_mass(indices)[0], 15.0 | units.kg)

        self.assertRaises(AmuseException, instance.get_mass, [4, 5],
            expected_message="Error when calling 'get_mass' of a '<class 'amuse.community.bhtree.interface.BHTree'>', errorcode is -1")

        instance.cleanup_code()
        instance.stop()

    def test7(self):
        convert_nbody = nbody_system.nbody_to_si(5.0 | units.kg, 10.0 | units.m)

        instance = BHTree(convert_nbody)
        instance.commit_parameters()

        particles = datamodel.Particles(2)
        self.assertEqual(len(instance.particles), 0)

        particles.mass = [15.0, 30.0] | units.kg
        particles.radius = [10.0, 20.0] | units.m
        particles.position = [[10.0, 20.0, 30.0], [20.0, 40.0, 60.0]] | units.m
        particles.velocity = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]] | units.m / units.s

        instance.particles.add_particles(particles)
        instance.commit_particles()

        self.assertEqual(instance.get_mass(1), 15.0 | units.kg)
        self.assertAlmostRelativeEquals(instance.get_position(1)[2], 30.0 | units.m)

        self.assertEqual(len(instance.particles), 2)

        self.assertAlmostRelativeEquals(instance.particles.mass[1], 30.0 | units.kg)
        self.assertAlmostRelativeEquals(instance.particles.position[1][2], 60.0 | units.m)
        instance.cleanup_code()
        instance.stop()

    def test8(self):
        convert_nbody = nbody_system.nbody_to_si(5.0 | units.kg, 10.0 | units.m)

        instance = BHTree(convert_nbody)
        instance.commit_parameters()

        particles = datamodel.Particles(2)
        self.assertEqual(len(instance.particles), 0)

        particles.mass = [15.0, 30.0] | units.kg
        particles.radius = [10.0, 20.0] | units.m
        particles.position = [[10.0, 20.0, 30.0], [20.0, 40.0, 60.0]] | units.m
        particles.velocity = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]] | units.m / units.s

        instance.particles.add_particles(particles)
        instance.commit_particles()

        instance.particles.mass = [17.0, 33.0] | units.kg

        self.assertEqual(instance.get_mass(1), 17.0 | units.kg)
        instance.cleanup_code()
        instance.stop()

    def test9(self):
        instance = BHTree()
        instance.initialize_code()
        instance.parameters.epsilon_squared = 0.00001 | nbody_system.length**2

        particles = datamodel.Particles(2)
        particles.mass = [1.0, 1.0] | nbody_system.mass
        particles.radius = [0.0001, 0.0001] | nbody_system.length
        particles.position = [[0.0, 0.0, 0.0], [2.0, 0.0, 0.0]] | nbody_system.length
        particles.velocity = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]] | nbody_system.speed
        instance.particles.add_particles(particles)

        zero = 0.0 | nbody_system.length
        fx, fy, fz = instance.get_gravity_at_point(zero, 1.0 | nbody_system.length, zero, zero)
        self.assertAlmostEqual(fx, 0.0 | nbody_system.acceleration, 3)
        self.assertAlmostEqual(fy, 0.0 | nbody_system.acceleration, 3)
        self.assertAlmostEqual(fz, 0.0 | nbody_system.acceleration, 3)

        for x in (0.25, 0.5, 0.75):
            x0 = x | nbody_system.length
            x1 = (2.0 - x) | nbody_system.length
            potential0 = instance.get_potential_at_point(zero, x0, zero, zero)
            potential1 = instance.get_potential_at_point(zero, x1, zero, zero)
            fx0, fy0, fz0 = instance.get_gravity_at_point(zero, x0, zero, zero)
            fx1, fy1, fz1 = instance.get_gravity_at_point(zero, x1, zero, zero)

            self.assertAlmostEqual(fy0, 0.0 | nbody_system.acceleration, 3)
            self.assertAlmostEqual(fz0, 0.0 | nbody_system.acceleration, 3)
            self.assertAlmostEqual(fy1, 0.0 | nbody_system.acceleration, 3)
            self.assertAlmostEqual(fz1, 0.0 | nbody_system.acceleration, 3)

            self.assertAlmostEqual(fx0, -1.0 * fx1, 5)
            fx = (-1.0 / (x0**2) + 1.0 / (x1**2)) * (1.0 | nbody_system.length ** 3 / nbody_system.time ** 2)
            self.assertAlmostEqual(fx, fx0, 2)
            self.assertAlmostEqual(potential0, potential1, 5)
        instance.cleanup_code()
        instance.stop()

    def test10(self):
        instance = BHTree()
        instance.initialize_code()
        instance.parameters.epsilon_squared = 0.00001 | nbody_system.length**2
        instance.commit_parameters()

        particles = datamodel.Particles(6)
        particles.mass = 1.0 | nbody_system.mass
        particles.radius = 0.00001 | nbody_system.length
        particles.position = [[-1.0, 0.0, 0.0], [1.0, 0.0, 0.0], [0.0, -1.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, -1.0], [0.0, 0.0, 1.0]] | nbody_system.length
        particles.velocity = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]] | nbody_system.speed
        instance.particles.add_particles(particles)
        instance.commit_particles()

        zero = 0.0 | nbody_system.length
        fx, fy, fz = instance.get_gravity_at_point(zero, zero, zero, zero)
        self.assertAlmostEqual(fx, 0.0 | nbody_system.acceleration, 3)
        self.assertAlmostEqual(fy, 0.0 | nbody_system.acceleration, 3)
        self.assertAlmostEqual(fz, 0.0 | nbody_system.acceleration, 3)

        for position in (0.25, 0.5, 0.75):
            p0 = position | nbody_system.length
            p1 = -position | nbody_system.length
            for i in range(3):
                args0 = [zero] * 4
                args1 = [zero] * 4
                args0[1 + i] = p0
                args1[1 + i] = p1
                f0 = instance.get_gravity_at_point(*args0)
                f1 = instance.get_gravity_at_point(*args1)

                for j in range(3):
                    if j != i:
                        self.assertAlmostEqual(f0[j], 0.0 | nbody_system.acceleration, 3)
                        self.assertAlmostEqual(f1[j], 0.0 | nbody_system.acceleration, 3)
                    else:
                        self.assertAlmostEqual(f0[j], -1.0 * f1[j], 5)

        instance.stop()

    def test11(self):

        convert_nbody = nbody_system.nbody_to_si(5.0 | units.kg, 10.0 | units.m)

        instance = BHTree(convert_nbody)

        particles = datamodel.Particles(2)
        self.assertEqual(len(instance.particles), 0)

        particles.mass = [15.0, 30.0] | units.kg
        particles.radius = [10.0, 20.0] | units.m
        particles.position = [[10.0, 20.0, 30.0], [20.0, 40.0, 60.0]] | units.m
        particles.velocity = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]] | units.m / units.s

        instance.particles.add_particles(particles)

        copyof = instance.particles.copy()

        self.assertAlmostEqual(30 | units.kg, copyof[1].mass, 6)

        copyof[1].mass = 35 | units.kg

        copyof.copy_values_of_all_attributes_to(instance.particles)

        self.assertAlmostEqual(35 | units.kg, instance.particles[1].mass, 6)
        instance.stop()

    def test12(self):

        convert_nbody = nbody_system.nbody_to_si(5.0 | units.kg, 10.0 | units.m)

        instance = BHTree(convert_nbody)
        instance.commit_parameters()

        particles = datamodel.Particles(2)
        self.assertEqual(len(instance.particles), 0)

        particles.mass = [15.0, 30.0] | units.kg
        particles.radius = [10.0, 20.0] | units.m
        particles.position = [[10.0, 20.0, 30.0], [20.0, 40.0, 60.0]] | units.m
        particles.velocity = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]] | units.m / units.s

        instance.particles.add_particles(particles)
        instance.commit_particles()

        copyof = instance.particles.copy()

        instance.set_state(1, 16 | units.kg, 20.0 | units.m, 40.0 | units.m, 60.0 | units.m,
                                 1.0 | units.ms, 1.0 | units.ms, 1.0 | units.ms)

        curr_state = instance.get_state(1)
        for expected, actual in zip((16 | units.kg, 20.0 | units.m, 40.0 | units.m, 60.0 | units.m,
                                 1.0 | units.ms, 1.0 | units.ms, 1.0 | units.ms, 0 | units.m), curr_state):
            self.assertAlmostRelativeEquals(actual, expected)

        instance.set_state(1, 16 | units.kg, 20.0 | units.m, 40.0 | units.m, 60.0 | units.m,
                                 1.0 | units.ms, 1.0 | units.ms, 1.0 | units.ms, 20.0 | units.m)

        curr_state = instance.get_state(1)
        for expected, actual in zip((16 | units.kg, 20.0 | units.m, 40.0 | units.m, 60.0 | units.m,
                                 1.0 | units.ms, 1.0 | units.ms, 1.0 | units.ms, 20 | units.m), curr_state):
            self.assertAlmostRelativeEquals(actual, expected)

        instance.stop()

    def test13(self):

        convert_nbody = nbody_system.nbody_to_si(1.0 | units.kg, 1.0 | units.m)

        instance = BHTree(convert_nbody)
        instance.commit_parameters()

        particles = datamodel.Particles(2)
        self.assertEqual(len(instance.particles), 0)

        particles.mass = [30.0, 30.0] | units.kg
        particles.radius = [1.0, 1.0] | units.m
        particles.position = [[-10.0, 0.0, 0.0], [10.0, 0.0, 0.0]] | units.m
        particles.velocity = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]] | units.m / units.s

        instance.particles.add_particles(particles)
        instance.commit_particles()

        copyof = instance.particles.copy()

        com = instance.center_of_mass_position
        self.assertAlmostEqual(com[0], quantities.new_quantity(0.0, units.m), constants.precision)
        instance.stop()

    def test14(self):
        print("Test14: Testing BHTree parameters (I)")
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.yr, 1.0 | units.AU)
        instance = BHTree(convert_nbody)

        value, error = instance.legacy_interface.get_epsilon_squared()
        self.assertEqual(0, error)
        self.assertEqual(0.125, value)
        self.assertAlmostEqual(0.125 | units.AU**2, instance.parameters.epsilon_squared, in_units=units.AU**2)
        for x in [0.01, 0.1, 0.2]:
            instance.parameters.epsilon_squared = x | units.AU**2
            self.assertAlmostEqual(x | units.AU**2, instance.parameters.epsilon_squared, in_units=units.AU**2)

        (value, error) = instance.legacy_interface.get_time_step()
        self.assertEqual(0, error)
        self.assertEqual(0.015625, value)
        self.assertAlmostEqual(0.015625 | units.yr, instance.parameters.timestep, in_units=units.yr)
        for x in [0.001, 0.01, 0.1]:
            instance.parameters.timestep = x | units.yr
            self.assertAlmostEqual(x | units.yr, instance.parameters.timestep, in_units=units.yr)

        (value, error) = instance.legacy_interface.get_theta_for_tree()
        self.assertEqual(0, error)
        self.assertEqual(0.75, value)
        self.assertEqual(0.75, instance.parameters.opening_angle)
        for x in [0.2, 0.5, 0.7]:
            instance.parameters.opening_angle = x
            self.assertEqual(x, instance.parameters.opening_angle)

        (value, error) = instance.legacy_interface.get_use_self_gravity()
        self.assertEqual(0, error)
        self.assertEqual(1, value)
        self.assertEqual(1, instance.parameters.use_self_gravity)
        for x in [0, 1]:
            instance.parameters.use_self_gravity = x
            self.assertEqual(x, instance.parameters.use_self_gravity)

        (value, error) = instance.legacy_interface.get_ncrit_for_tree()
        self.assertEqual(0, error)
        self.assertEqual(12, value)
        self.assertEqual(12, instance.parameters.ncrit_for_tree)
        for x in [512, 2048, 4096]:
            instance.parameters.ncrit_for_tree = x
            self.assertEqual(x, instance.parameters.ncrit_for_tree)

        (value, error) = instance.legacy_interface.get_dt_dia()
        self.assertEqual(0, error)
        self.assertEqual(1.0, value)
        self.assertAlmostEqual(1.0 | units.yr, instance.parameters.dt_dia, in_units=units.yr)
        for x in [0.1, 10.0, 100.0]:
            instance.parameters.dt_dia = x | units.yr
            self.assertAlmostEqual(x | units.yr, instance.parameters.dt_dia, in_units=units.yr)
        instance.stop()

    def test15(self):
        print("Test15: Testing effect of BHTree parameter epsilon_squared")
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.MSun, 1.0 | units.AU)

        particles = datamodel.Particles(2)
        sun = particles[0]
        sun.mass = 1.0 | units.MSun
        sun.position = [0.0, 0.0, 0.0] | units.AU
        sun.velocity = [0.0, 0.0, 0.0] | units.AU / units.yr
        sun.radius = 1.0 | units.RSun

        earth = particles[1]
        earth.mass = 5.9736e24 | units.kg
        earth.radius = 6371.0 | units.km
        earth.position = [0.0, 1.0, 0.0] | units.AU
        earth.velocity = [2.0*numpy.pi, -0.0001, 0.0] | units.AU / units.yr

        initial_direction = math.atan((earth.velocity[0]/earth.velocity[1]))
        final_direction = []
        for log_eps2 in range(-9, 10, 2):
            instance = BHTree(convert_nbody)
            instance.initialize_code()
            instance.parameters.epsilon_squared = 10.0**log_eps2 | units.AU ** 2
            instance.particles.add_particles(particles)
            instance.commit_particles()
            instance.evolve_model(0.25 | units.yr)
            final_direction.append(math.atan((instance.particles[1].velocity[0] /
                instance.particles[1].velocity[1])))
            instance.stop()
        # Small values of epsilon_squared should result in normal earth-sun dynamics: rotation of 90 degrees
        self.assertAlmostEqual(abs(final_direction[0]), abs(initial_direction+math.pi/2.0), 2)
        # Large values of epsilon_squared should result in ~ no interaction
        self.assertAlmostEqual(final_direction[-1], initial_direction, 2)
        # Outcome is most sensitive to epsilon_squared when epsilon_squared = d(earth, sun)^2
        delta = [abs(final_direction[i+1]-final_direction[i]) for i in range(len(final_direction)-1)]
        self.assertEqual(delta[len(final_direction)//2 - 1], max(delta))

    def test16(self):
        numpy.random.seed(0)
        number_of_stars = 2
        stars = plummer.new_plummer_model(number_of_stars)
        stars.radius = 0.00001 | nbody_system.length
        stars.scale_to_standard()

        instance = BHTree()
        instance.initialize_code()
        instance.parameters.epsilon_squared = (1.0 / 20.0 / (number_of_stars**0.33333) | nbody_system.length)**2
        instance.parameters.timestep = 0.004 | nbody_system.time
        instance.parameters.timestep = 0.00001 | nbody_system.time
        instance.commit_parameters()
        print(instance.parameters.timestep)
        instance.particles.add_particles(stars)
        instance.commit_particles()
        energy_total_t0 = instance.potential_energy + instance.kinetic_energy
        request = instance.evolve_model.asynchronous(1.0 | nbody_system.time)
        request.result()
        energy_total_t1 = instance.potential_energy + instance.kinetic_energy

        self.assertAlmostRelativeEqual(energy_total_t0, energy_total_t1, 3)
        instance.stop()
        numpy.random.seed()

    def test17(self):
        print("Testing BHTree collision_detection")
        particles = datamodel.Particles(7)
        particles.mass = 0.001 | nbody_system.mass
        particles.radius = 0.01 | nbody_system.length
        particles.x = [-101.0, -100.0, -0.5, 0.5, 100.0, 101.0, 104.0] | nbody_system.length
        particles.y = 0 | nbody_system.length
        particles.z = 0 | nbody_system.length
        particles.velocity = [[2, 0, 0], [-2, 0, 0]]*3 + [[-4, 0, 0]] | nbody_system.speed

        instance = BHTree(redirection='none')
        instance.initialize_code()
        instance.parameters.set_defaults()

        # Uncommenting any of the following two lines will suppress collision detection
# ~        instance.parameters.use_self_gravity = 0
# ~        instance.parameters.epsilon_squared = 0.0 | nbody_system.length**2

        instance.parameters.opening_angle = 0.1
        instance.particles.add_particles(particles)
        collisions = instance.stopping_conditions.collision_detection
        collisions.enable()
        instance.evolve_model(1.0 | nbody_system.time)

        self.assertTrue(collisions.is_set())
        self.assertTrue(instance.model_time < 0.5 | nbody_system.time)
        self.assertEqual(len(collisions.particles(0)), 3)
        self.assertEqual(len(collisions.particles(1)), 3)
        self.assertEqual(len(particles - collisions.particles(0) - collisions.particles(1)), 1)
        self.assertEqual(abs(collisions.particles(0).x - collisions.particles(1).x) <
                (collisions.particles(0).radius + collisions.particles(1).radius),
                [True, True, True])

        sticky_merged = datamodel.Particles(len(collisions.particles(0)))
        sticky_merged.mass = collisions.particles(0).mass + collisions.particles(1).mass
        sticky_merged.radius = collisions.particles(0).radius
        for p1, p2, merged in zip(collisions.particles(0), collisions.particles(1), sticky_merged):
            merged.position = (p1 + p2).center_of_mass()
            merged.velocity = (p1 + p2).center_of_mass_velocity()

        print(instance.model_time)
        print(instance.particles)
        instance.particles.remove_particles(collisions.particles(0) + collisions.particles(1))
        instance.particles.add_particles(sticky_merged)

        instance.evolve_model(1.0 | nbody_system.time)
        print()
        print(instance.model_time)
        print(instance.particles)
        self.assertTrue(collisions.is_set())
        self.assertTrue(instance.model_time < 1.0 | nbody_system.time)
        self.assertEqual(len(collisions.particles(0)), 1)
        self.assertEqual(len(collisions.particles(1)), 1)
        self.assertEqual(len(instance.particles - collisions.particles(0) - collisions.particles(1)), 2)
        self.assertEqual(abs(collisions.particles(0).x - collisions.particles(1).x) <
                (collisions.particles(0).radius + collisions.particles(1).radius),
                [True])
        instance.stop()

    def test18(self):
        particles = datamodel.Particles(2)
        particles.x = [0.0, 10.0] | nbody_system.length
        particles.y = 0 | nbody_system.length
        particles.z = 0 | nbody_system.length
        particles.radius = 0.005 | nbody_system.length
        particles.vx = 0 | nbody_system.speed
        particles.vy = 0 | nbody_system.speed
        particles.vz = 0 | nbody_system.speed
        particles.mass = 1.0 | nbody_system.mass

        instance = BHTree()
        instance.initialize_code()
        instance.parameters.stopping_conditions_number_of_steps = 2
        self.assertEqual(instance.parameters.stopping_conditions_number_of_steps, 2)
        instance.parameters.epsilon_squared = (0.01 | nbody_system.length)**2
        instance.particles.add_particles(particles)
        instance.stopping_conditions.number_of_steps_detection.enable()
        instance.evolve_model(10 | nbody_system.time)
        self.assertTrue(instance.stopping_conditions.number_of_steps_detection.is_set())
        self.assertTrue(instance.model_time < 10 | nbody_system.time)
        instance.stop()

    def test19(self):
        particles = datamodel.Particles(2)
        particles.x = [0.0, 10.0] | nbody_system.length
        particles.y = 0.0 | nbody_system.length
        particles.z = 0.0 | nbody_system.length
        particles.radius = 0.005 | nbody_system.length
        particles.vx = 0.0 | nbody_system.speed
        particles.vy = 0.0 | nbody_system.speed
        particles.vz = 0.0 | nbody_system.speed
        particles.mass = 1.0 | nbody_system.mass

        very_short_time_to_evolve = 1 | units.s
        very_long_time_to_evolve = 1e9 | nbody_system.time

        instance = BHTree()
        instance.initialize_code()
        instance.parameters.stopping_conditions_timeout = very_short_time_to_evolve
        self.assertEqual(instance.parameters.stopping_conditions_timeout, very_short_time_to_evolve)
        instance.parameters.epsilon_squared = (0.01 | nbody_system.length)**2
        instance.particles.add_particles(particles)
        instance.stopping_conditions.timeout_detection.enable()
        start = time.time()
        instance.evolve_model(very_long_time_to_evolve)
        end = time.time()
        self.assertTrue(instance.stopping_conditions.timeout_detection.is_set())
        self.assertTrue((end-start) < very_short_time_to_evolve.value_in(units.s) + 2)  # 2 = some overhead compensation
        instance.stop()

    def test20(self):
        particles = datamodel.Particles(2)
        particles.x = [0.0, 10.0] | nbody_system.length
        particles.y = 0.0 | nbody_system.length
        particles.z = 0.0 | nbody_system.length
        particles.radius = 0.005 | nbody_system.length
        particles.vx = 0.0 | nbody_system.speed
        particles.vy = 0.0 | nbody_system.speed
        particles.vz = 0.0 | nbody_system.speed
        particles.mass = 1.0 | nbody_system.mass

        very_short_time_to_evolve = 1 | units.s
        very_long_time_to_evolve = 1e9 | nbody_system.time

        instance = BHTree()
        instance.initialize_code()
        instance.parameters.stopping_conditions_timeout = very_short_time_to_evolve
        self.assertEqual(instance.parameters.stopping_conditions_timeout, very_short_time_to_evolve)
        instance.parameters.epsilon_squared = (0.01 | nbody_system.length)**2
        instance.particles.add_particles(particles)
        codeparticles1 = instance.particles
        instance.particles.add_particle(datamodel.Particle(
            position=[0, 1, 2] | nbody_system.length,
            velocity=[0, 0, 0] | nbody_system.speed,
            radius=0.005 | nbody_system.length,
            mass=1 | nbody_system.mass
        ))
        codeparticles2 = instance.particles
        self.assertTrue(codeparticles1 is codeparticles2)
        instance.cleanup_code()
        codeparticles3 = instance.particles
        self.assertFalse(codeparticles1 is codeparticles3)

        instance.stop()

    def test21(self):
        particles = datamodel.Particles(2)
        particles.x = [0.0, 10.0] | nbody_system.length
        particles.y = 0.0 | nbody_system.length
        particles.z = 0.0 | nbody_system.length
        particles.radius = 0.005 | nbody_system.length
        particles.vx = 0.0 | nbody_system.speed
        particles.vy = 0.0 | nbody_system.speed
        particles.vz = 0.0 | nbody_system.speed
        particles.mass = 1.0 | nbody_system.mass

        very_short_time_to_evolve = 1 | units.s
        very_long_time_to_evolve = 1e9 | nbody_system.time

        instance = BHTree()
        instance.initialize_code()
        instance.parameters.epsilon_squared = (1e-5 | nbody_system.length)**2
        instance.particles.add_particles(particles)
        instance.commit_particles()
        self.assertAlmostRelativeEquals(instance.potential_energy, -0.1 | nbody_system.energy, 5)
        instance.stop()

    def test22(self):
        particles = datamodel.Particles(2)
        particles.x = [0.0, 10.0] | nbody_system.length
        particles.y = 0.0 | nbody_system.length
        particles.z = 0.0 | nbody_system.length
        particles.vx = 0.0 | nbody_system.speed
        particles.vy = 0.0 | nbody_system.speed
        particles.vz = 0.0 | nbody_system.speed
        particles.mass = 1.0 | nbody_system.mass

        instance = BHTree()
        instance.particles.add_particles(particles)
        instance.commit_particles()
        self.assertEqual(instance.particles[0].radius, 0.0 | nbody_system.length)
        p = datamodel.Particle(
            x=1.0 | nbody_system.length,
            y=2.0 | nbody_system.length,
            z=3.0 | nbody_system.length,
            vx=1.0 | nbody_system.speed,
            vy=2.0 | nbody_system.speed,
            vz=3.0 | nbody_system.speed,
            mass=1.0 | nbody_system.mass,
            radius=4.0 | nbody_system.length,
        )
        instance.particles.add_particle(p)
        self.assertEqual(instance.particles[0].radius, 0.0 | nbody_system.length)
        self.assertEqual(instance.particles[1].radius, 0.0 | nbody_system.length)
        self.assertEqual(instance.particles[2].radius, 4.0 | nbody_system.length)

        instance.stop()

    def test23(self):
        particles = datamodel.Particles(2)
        particles.x = [0.0, 10.0] | nbody_system.length
        particles.y = 0.0 | nbody_system.length
        particles.z = 0.0 | nbody_system.length
        particles.vx = 1.0 | nbody_system.speed
        particles.vy = 0.0 | nbody_system.speed
        particles.vz = 0.0 | nbody_system.speed
        particles.mass = 0.1 | nbody_system.mass

        instance = BHTree(redirection="none")
        instance.particles.add_particles(particles)
        instance.commit_particles()
        instance.evolve_model(0.1 | nbody_system.time)
        self.assertFalse(instance.particles[0].vy > 0 | nbody_system.speed)
        self.assertAlmostRelativeEquals(instance.particles[0].x, 0.1 | nbody_system.length, 4)
        instance.particles.new_channel_to(particles).copy()
        particles.vy = 1 | nbody_system.speed
        particles.new_channel_to(instance.particles).copy()

        instance.evolve_model(0.2 | nbody_system.time)
        self.assertTrue(instance.particles[0].vy > 0 | nbody_system.speed)
        self.assertAlmostRelativeEquals(instance.particles[0].y, 0.1 | nbody_system.length, 4)
        instance.stop()



import numpy

from amuse.community.twobody import interface

from amuse.test.amusetest import TestWithMPI
from amuse.units import units
from amuse.units import nbody_system
from amuse.units.quantities import zero
from amuse.support.exceptions import AmuseException
from amuse import datamodel


class TwoBodyCodeTests(TestWithMPI):

    def test_stumpff(self):
        self.assertAlmostEqual(interface.stumpff_C(0), interface.stumpff_C(0.0001), 5)
        self.assertAlmostEqual(interface.stumpff_C(0), interface.stumpff_C(-0.0001), 5)
        self.assertAlmostEqual(interface.stumpff_S(0), interface.stumpff_S(0.0001), 5)
        self.assertAlmostEqual(interface.stumpff_S(0), interface.stumpff_S(-0.0001), 5)


class TwoBodyInterfaceTests(TestWithMPI):

    def test1(self):

        instance = interface.TwoBodyInterface()

        res1 = instance.new_particle(mass=11.0, radius=2.0, x=0.0, y=0.0, z=0.0, vx=0.0, vy=0.0, vz=0.0)
        res2 = instance.new_particle(mass=21.0, radius=5.0, x=10.0, y=0.0, z=0.0, vx=10.0, vy=0.0, vz=0.0)

        self.assertEqual(0, res1['index_of_the_particle'])
        self.assertEqual(1, res2['index_of_the_particle'])

        retrieved_state1 = instance.get_state(0)
        retrieved_state2 = instance.get_state(1)

        self.assertEqual(11.0,  retrieved_state1['mass'])
        self.assertEqual(21.0,  retrieved_state2['mass'])
        self.assertEqual(0.0,  retrieved_state1['x'])
        self.assertEqual(10.0,  retrieved_state2['x'])

        instance.stop()

    def test2(self):

        instance = interface.TwoBodyInterface()

        res1 = instance.new_particle(mass=10.0, radius=0.0, x=-1.0, y=0.0, z=0.0, vx=0.0, vy=10.0, vz=0.0)
        res2 = instance.new_particle(mass=10.0, radius=0.0, x=1.0, y=0.0, z=0.0, vx=0.0, vy=-10.0, vz=0.0)

        ek = 0.5*(10*100+10*100)
        ep = -(10*10/2)

        e, err = instance.get_kinetic_energy()
        self.assertEqual(ek, e)

        e, err = instance.get_potential_energy()
        self.assertEqual(ep, e)

        instance.stop()


class TwoBodyTests(TestWithMPI):

    def test1(self):
        convert_nbody = nbody_system.nbody_to_si(5.9742e24 | units.kg, 1e6 | units.m)
        instance = interface.TwoBody(convert_nbody)
        instance.stop()

    def test2(self):
        convert_nbody = nbody_system.nbody_to_si(5.9742e24 | units.kg, 1e6 | units.m)
        instance = interface.TwoBody(convert_nbody)

        p = datamodel.Particle()
        p.mass = 5.9742e24 | units.kg
        p.radius = 6.371e6 | units.m
        p.position = [0., 7.e6, -1.2124e7] | units.m
        p.velocity = [0., 2.6679e3, 4.6210e3] | units.m/units.s

        instance.particles.add_particle(p)

        instance.evolve_model(3600.0 | units.s)

        position = instance.particles[0].position
        velocity = instance.particles[0].velocity

        self.assertAlmostEqual(position.x.value_in(units.m), 0., 7)
        self.assertAlmostEqual(position.y.value_in(units.m)/(-3.30647600568e6), 1., 7)
        self.assertAlmostEqual(position.z.value_in(units.m)/7.40831575351e6, 1., 7)
        self.assertAlmostEqual(velocity.x.value_in(units.m / units.s), 0., 7)
        self.assertAlmostEqual(velocity.y.value_in(units.m / units.s)/(-8.29821376206e3), 1., 7)
        self.assertAlmostEqual(velocity.z.value_in(units.m / units.s)/(-0.972888312209e3), 1., 7)
        instance.stop()

    def test3(self):
        convert_nbody = nbody_system.nbody_to_si(5.9742e24 | units.kg, 1e6 | units.m)
        instance = interface.TwoBody(convert_nbody)

        p = datamodel.Particle()
        p.mass = 5.9742e24 | units.kg
        p.radius = 7.1e6 | units.m
        p.position = [0., 7.e6, -1.2124e7] | units.m
        p.velocity = [0., 2.6679e3, 4.6210e3] | units.m/units.s

        instance.particles.add_particle(p)

        instance.evolve_model(3600.0 | units.s)

        dt = convert_nbody.to_si(instance.model_time)
        self.assertAlmostEqual(dt.value_in(units.s)/2583.44780926, 1., 7)

        position = instance.particles[0].position
        self.assertAlmostEqual(((position.x**2+position.y**2+position.z**2)/(7.1e6)**2).value_in(units.m**2), 1., 7)
        instance.stop()

    def test4(self):
        convert_nbody = nbody_system.nbody_to_si(5.9742e24 | units.kg, 1e6 | units.m)
        instance = interface.TwoBody(convert_nbody)

        p = datamodel.Particle()
        p.mass = 5.9742e24 | units.kg
        p.radius = 7.1e6 | units.m
        p.position = [0., 7.e6, -1.2124e7] | units.m
        p.velocity = [0., 2.6679e3, 4.6210e3] | units.m/units.s

        instance.particles.add_particle(p)

        instance.evolve_model(3600.0 | units.s)

        dt = convert_nbody.to_si(instance.model_time)

        self.assertEqual(instance.particles[0].mass, 5.9742e24 | units.kg)

        instance.particles[0].mass = 0.8*5.9742e24 | units.kg
        instance.evolve_model(4000.0 | units.s)
        self.assertEqual(instance.particles.mass[0], 0.8*5.9742e24 | units.kg)
        instance.stop()

    def test5(self):
        # from: Fundamentals of Celestial Mechanics, J.M.A. Danby 2nd edition
        instance = interface.TwoBody()
        p = datamodel.Particle()

        p.mass = 1.0 | nbody_system.mass
        p.radius = 0.001 | nbody_system.length
        p.position = [1.0, 0.1, -0.1] | nbody_system.length
        p.velocity = [-0.1, 2.0, -0.2] | nbody_system.speed

        instance.particles.add_particle(p)
        instance.evolve_model(1.0 | nbody_system.time)
        self.assertAlmostEqual(instance.particles.x, 0.611238439231 | nbody_system.length, 7)
        self.assertAlmostEqual(instance.particles.y, 1.92873971354574 | nbody_system.length, 7)
        self.assertAlmostEqual(instance.particles.z, -0.2562478900031234 | nbody_system.length, 7)
        instance.stop()

    def test6(self):
        # from: Fundamentals of Celestial Mechanics, J.M.A. Danby 2nd edition
        instance = interface.TwoBody()
        p = datamodel.Particles(2)

        p.mass = [1, 0.0] | nbody_system.mass
        p.radius = 0.001 | nbody_system.length
        p.x = [1.0,  0.0] | nbody_system.length
        p.y = [0.1,  0.0] | nbody_system.length
        p.z = [-0.1, 0.0] | nbody_system.length

        p.vx = [-0.1, 0.0] | nbody_system.speed
        p.vy = [2.0, 0.0] | nbody_system.speed
        p.vz = [-0.2, 0.0] | nbody_system.speed

        instance.particles.add_particles(p)
        instance.evolve_model(1.0 | nbody_system.time)
        self.assertAlmostEqual(instance.particles.x[0] - instance.particles.x[1], 0.611238439231 | nbody_system.length, 7)
        self.assertAlmostEqual(instance.particles.y[0] - instance.particles.y[1], 1.92873971354574 | nbody_system.length, 7)
        self.assertAlmostEqual(instance.particles.z[0] - instance.particles.z[1], -0.2562478900031234 | nbody_system.length, 7)
        instance.stop()

    def test7(self):
        print("Test 7: get_gravity_at_point")
        instance = interface.TwoBody()
        p = datamodel.Particles(2)

        p.mass = 0.5 | nbody_system.mass
        p.radius = 0 | nbody_system.length
        p.position = [[1.0, 0, 0], [-1.0, 0, 0]] | nbody_system.length
        p.velocity = [[0]*3]*2 | nbody_system.speed

        instance.particles.add_particles(p)
        zero = [0.0] | nbody_system.length
        ax, ay, az = instance.get_gravity_at_point(zero, 0.5 | nbody_system.length, zero, zero)
        self.assertAlmostEqual(ax, 16/9.0 | nbody_system.acceleration)
        self.assertAlmostEqual(ay, 0 | nbody_system.acceleration)
        self.assertAlmostEqual(az, 0 | nbody_system.acceleration)

        zero = [0.0]*4 | nbody_system.length
        ax, ay, az = instance.get_gravity_at_point(zero, [0, 0.5, 1.5, 2.5] | nbody_system.length, zero, zero)
        self.assertAlmostEqual(ax, [0, 16/9.0, -2.08, -0.26303854] | nbody_system.acceleration)
        self.assertAlmostEqual(ay, 0 | nbody_system.acceleration)
        self.assertAlmostEqual(az, 0 | nbody_system.acceleration)

        zero = [0.0] | nbody_system.length
        ax, ay, az = instance.get_gravity_at_point(zero, 1.0 | nbody_system.length, zero, zero)
        self.assertTrue(numpy.isnan(ax[0].value_in(nbody_system.acceleration)))
        self.assertTrue(numpy.isnan(ay[0].value_in(nbody_system.acceleration)))
        self.assertTrue(numpy.isnan(az[0].value_in(nbody_system.acceleration)))

        zero = [0.0] | nbody_system.length
        eps = [0.1] | nbody_system.length
        ax, ay, az = instance.get_gravity_at_point(eps, 1.0 | nbody_system.length, zero, zero)
        self.assertAlmostEqual(ax, -0.12453271 | nbody_system.acceleration)
        self.assertAlmostEqual(ay, 0 | nbody_system.acceleration)
        self.assertAlmostEqual(az, 0 | nbody_system.acceleration)
        instance.stop()

    def test8(self):
        print("Test 8: get_potential_at_point")
        instance = interface.TwoBody()
        p = datamodel.Particles(2)

        p.mass = 0.5 | nbody_system.mass
        p.radius = 0 | nbody_system.length
        p.position = [[1.0, 0, 0], [-1.0, 0, 0]] | nbody_system.length
        p.velocity = [[0]*3]*2 | nbody_system.speed

        instance.particles.add_particles(p)
        zero = [0.0] | nbody_system.length
        phi = instance.get_potential_at_point(zero, 0.5 | nbody_system.length, zero, zero)
        self.assertAlmostEqual(phi, -4/3.0 | nbody_system.potential)

        zero = [0.0]*4 | nbody_system.length
        phi = instance.get_potential_at_point(zero, [0, 0.5, 1.5, 2.5] | nbody_system.length, zero, zero)
        self.assertAlmostEqual(phi, [-1, -4/3.0, -1.2, -0.47619047] | nbody_system.potential)

        zero = [0.0] | nbody_system.length
        phi = instance.get_potential_at_point(zero, 1.0 | nbody_system.length, zero, zero)
        self.assertTrue(numpy.isinf(phi[0].value_in(nbody_system.potential)))
        instance.stop()

    def test9(self):
        print("Test 9: TwoBody parameters")
        instance = interface.TwoBody()

        self.assertEqual(instance.parameters.epsilon_squared, zero)

        self.assertRaises(AmuseException, setattr, instance.parameters, "epsilon_squared", zero,
            expected_message="Could not set value for parameter 'epsilon_squared' of a 'TwoBody' object, "
                "parameter is read-only")

        instance.stop()

    def test10(self):
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.yr, 1.0 | units.AU)
        instance = interface.TwoBody(convert_nbody)

        value = instance.get_begin_time()
        self.assertEqual(0.0 | units.yr, value)
        self.assertAlmostEqual(0.0 | units.yr, instance.parameters.begin_time, in_units=units.yr)
        for x in [1.0, 10.0, 100.0]:
            instance.parameters.begin_time = x | units.yr
            self.assertAlmostEqual(x | units.yr, instance.parameters.begin_time, in_units=units.yr)
        instance.stop()

    def test11(self):
        instance = interface.TwoBody()
        p = datamodel.Particles(2)

        p.mass = [1, 0.0] | nbody_system.mass
        p.radius = 0.001 | nbody_system.length
        p.x = [1.0,  0.0] | nbody_system.length
        p.y = [0.1,  0.0] | nbody_system.length
        p.z = [-0.1, 0.0] | nbody_system.length

        p.vx = [-0.1, 0.0] | nbody_system.speed
        p.vy = [2.0, 0.0] | nbody_system.speed
        p.vz = [-0.2, 0.0] | nbody_system.speed

        instance.particles.add_particles(p)
        instance.evolve_model(0.5 | nbody_system.time)
        particles1 = instance.particles.copy()
        instance.stop()

        instance = interface.TwoBody()
        instance.parameters.begin_time = 0.5 | nbody_system.time
        instance.particles.add_particles(particles1)
        instance.evolve_model(1.0 | nbody_system.time)

        self.assertAlmostEqual(instance.particles.x[0] - instance.particles.x[1], 0.611238439231 | nbody_system.length, 7)
        self.assertAlmostEqual(instance.particles.y[0] - instance.particles.y[1], 1.92873971354574 | nbody_system.length, 7)
        self.assertAlmostEqual(instance.particles.z[0] - instance.particles.z[1], -0.2562478900031234 | nbody_system.length, 7)
        instance.stop()



import numpy
from amuse.test.amusetest import TestWithMPI

from amuse.support.exceptions import AmuseException
from amuse.ext.sph_to_grid import convert_SPH_to_grid
from amuse.units import units, generic_unit_system, constants
from amuse.units.generic_unit_converter import ConvertBetweenGenericAndSiUnits
from amuse.datamodel import Particles
from amuse.community.gadget2.interface import Gadget2
from amuse.community.fi.interface import Fi


class TestSPH2Grid(TestWithMPI):

    def setup_sph_code(self, sph_code, number_of_particles, L, rho, u):
        converter = ConvertBetweenGenericAndSiUnits(L, rho, constants.G)
        sph_code = sph_code(converter, mode='periodic')  # , redirection = 'none')
        sph_code.parameters.periodic_box_size = 10.0 | units.parsec
        gas = Particles(number_of_particles)
        gas.mass = (rho * L**3) / number_of_particles
        numpy.random.seed(12345)
        gas.x = L * numpy.random.uniform(0.0, 1.0, number_of_particles)
        gas.y = L * numpy.random.uniform(0.0, 1.0, number_of_particles)
        gas.z = L * numpy.random.uniform(0.0, 1.0, number_of_particles)
        gas.vx = numpy.zeros(number_of_particles) | units.cm / units.s
        gas.vy = numpy.zeros(number_of_particles) | units.cm / units.s
        gas.vz = numpy.zeros(number_of_particles) | units.cm / units.s
        gas.u = u
        if isinstance(sph_code, Fi):
            sph_code.parameters.self_gravity_flag = False
            sph_code.parameters.timestep = 0.1 | generic_unit_system.time
            gas.h_smooth = L / number_of_particles**(1/3.0)
            gas.position -= 0.5 * L

        sph_code.gas_particles.add_particles(gas)
        sph_code.commit_particles()
# ~        sph_code.evolve_model(0.01 | units.Myr)
        return sph_code

    def test1(self):
        print("Testing convert_SPH_to_grid with Gadget2")
        number_of_particles = 10000
        L = 10.0 | units.parsec
        rho = 1.14 | units.amu/units.cm**3
        u = 5.e11 | units.cm**2 / units.s**2

        sph_code = self.setup_sph_code(Gadget2, number_of_particles, L, rho, u)
        grid = convert_SPH_to_grid(sph_code, (10, 10, 10))
        sph_code.stop()

        self.assertEqual(grid.shape, (10, 10, 10))
        self.assertAlmostEqual(grid.cellsize(), (L/10).as_vector_with_length(3))
        self.assertAlmostEqual(grid[0, 0, 0].position, (L/20).as_vector_with_length(3))
        self.assertEqual(
            grid.contains([[x, x, x] for x in numpy.linspace(-1.0, 11.0, 4)] | units.parsec),
            [False, True, True, False])
        self.assertAlmostEqual(grid.get_volume(), 1000.0 | units.parsec**3)

        mean_rho = grid.rho.mean()
        print("Mean density:", mean_rho.as_quantity_in(units.amu/units.cm**3), " -  original:", rho)
        self.assertAlmostRelativeEqual(mean_rho, rho, 1)
        self.assertAlmostRelativeEqual(grid.energy.mean(), mean_rho * u, 1)
        self.assertEqual(grid.momentum, 0 * mean_rho * u.sqrt())

    def test2(self):
        print("Testing convert_SPH_to_grid with Fi")
        number_of_particles = 10000
        L = 10.0 | units.parsec
        rho = 1.14 | units.amu/units.cm**3
        u = 5.e11 | units.cm**2 / units.s**2

        sph_code = self.setup_sph_code(Fi, number_of_particles, L, rho, u)
        grid = convert_SPH_to_grid(sph_code, (10, 10, 10))
        sph_code.stop()

        self.assertEqual(grid.shape, (10, 10, 10))
        self.assertAlmostEqual(grid.cellsize(), (L/10).as_vector_with_length(3))
        self.assertAlmostEqual(grid[0, 0, 0].position, (L/20).as_vector_with_length(3))
        self.assertEqual(
            grid.contains([[x, x, x] for x in numpy.linspace(-1.0, 11.0, 4)] | units.parsec),
            [False, True, True, False])
        self.assertAlmostEqual(grid.get_volume(), 1000.0 | units.parsec**3)

        mean_rho = grid.rho.mean()
        print("Mean density:", mean_rho.as_quantity_in(units.amu/units.cm**3), " -  original:", rho)
        self.assertAlmostRelativeEqual(mean_rho, rho, 1)
        self.assertAlmostRelativeEqual(grid.energy.mean(), mean_rho * u, 1)
        self.assertEqual(grid.momentum, 0 * mean_rho * u.sqrt())

    def test3(self):
        print("Testing convert_SPH_to_grid with Gadget2 and do_scale")
        number_of_particles = 10000
        L = 10.0 | units.parsec
        rho = 1.14 | units.amu/units.cm**3
        u = 5.e11 | units.cm**2 / units.s**2

        sph_code = self.setup_sph_code(Gadget2, number_of_particles, L, rho, u)
        grid = convert_SPH_to_grid(sph_code, (10, 10, 10), do_scale=True)
        sph_code.stop()

        self.assertEqual(grid.shape, (10, 10, 10))
        self.assertAlmostEqual(grid.cellsize(), (L/10).as_vector_with_length(3))
        self.assertAlmostEqual(grid[0, 0, 0].position, (L/20).as_vector_with_length(3))
        self.assertEqual(
            grid.contains([[x, x, x] for x in numpy.linspace(-1.0, 11.0, 4)] | units.parsec),
            [False, True, True, False])
        self.assertAlmostEqual(grid.get_volume(), 1000.0 | units.parsec**3)

        mean_rho = grid.rho.mean()
        print("Mean density:", mean_rho.as_quantity_in(units.amu/units.cm**3), " -  original:", rho)
        self.assertAlmostRelativeEqual(mean_rho, rho, 7)
        self.assertAlmostRelativeEqual(grid.energy.mean(), mean_rho * u, 7)
        self.assertEqual(grid.momentum, 0 * mean_rho * u.sqrt())

    def test4(self):
        print("Testing convert_SPH_to_grid with Fi and do_scale")
        number_of_particles = 10000
        L = 10.0 | units.parsec
        rho = 1.14 | units.amu/units.cm**3
        u = 5.e11 | units.cm**2 / units.s**2

        sph_code = self.setup_sph_code(Fi, number_of_particles, L, rho, u)
        grid = convert_SPH_to_grid(sph_code, (10, 10, 10), do_scale=True)
        sph_code.stop()

        self.assertEqual(grid.shape, (10, 10, 10))
        self.assertAlmostEqual(grid.cellsize(), (L/10).as_vector_with_length(3))
        self.assertAlmostEqual(grid[0, 0, 0].position, (L/20).as_vector_with_length(3))
        self.assertEqual(
            grid.contains([[x, x, x] for x in numpy.linspace(-1.0, 11.0, 4)] | units.parsec),
            [False, True, True, False])
        self.assertAlmostEqual(grid.get_volume(), 1000.0 | units.parsec**3)

        mean_rho = grid.rho.mean()
        print("Mean density:", mean_rho.as_quantity_in(units.amu/units.cm**3), " -  original:", rho)
        self.assertAlmostRelativeEqual(mean_rho, rho, 7)
        self.assertAlmostRelativeEqual(grid.energy.mean(), mean_rho * u, 7)
        self.assertEqual(grid.momentum, 0 * mean_rho * u.sqrt())

    def test5(self):
        print("Testing exceptions")

        class BogusHydroCode(object):
            MODE_PERIODIC_BOUNDARIES = "periodic"

            def __init__(self, mode="normal"):
                self.mode = mode

        self.assertRaises(AmuseException, convert_SPH_to_grid, BogusHydroCode(), (10, 10, 10),
            expected_message="Only periodic boundary conditions supported")

        self.assertRaises(AmuseException, convert_SPH_to_grid, BogusHydroCode(mode='periodic'), (10, 10),
            expected_message="Argument dimensions must contain exactly three numbers")

        self.assertRaises(AmuseException, convert_SPH_to_grid, BogusHydroCode(mode='periodic'), (10, 10, 10),
            expected_message="Unknown hydrodynamics code: BogusHydroCode - don't know whether the "
            "box runs from 0 to L or from -0.5 L to 0.5 L.")
